"""Group containing all static aero calculations.

Flaps modeling occurs as a static calculation to provide CL max and CL/CD increments to
the dynamic aero.
"""

import openmdao.api as om
from dymos.models.atmosphere.atmos_1976 import USatm1976Comp

from aviary.subsystems.aerodynamics.gasp_based.flaps_model import FlapsGroup
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic, Mission

# TODO: add subsystems to compute CLMXFU, CLMXTO, CLMXLD using dynamic aero components
# with alpha > alpha_stall


class PreMissionAero(om.Group):
    """Takeoff and landing flaps modeling"""

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        self.add_subsystem(
            "atmos",
            USatm1976Comp(
                num_nodes=1),
            promotes_inputs=[
                ("h",
                 "alt_flaps")],
            promotes_outputs=[
                ("temp",
                 Dynamic.Mission.TEMPERATURE),
                ("pres",
                 Dynamic.Mission.STATIC_PRESSURE),
                ("sos",
                 Dynamic.Mission.SPEED_OF_SOUND),
                "rho",
                "viscosity"],
        )

        self.add_subsystem(
            "kin_visc",
            om.ExecComp(
                "kinematic_viscosity = viscosity / rho",
                viscosity={"units": "lbf*s/ft**2"},
                rho={"units": "slug/ft**3"},
                kinematic_viscosity={"units": "ft**2/s"},
            ),
            promotes=["*"],
        )

        self.add_subsystem(
            "flaps_up",
            FlapsGroup(aviary_options=aviary_options),
            promotes_inputs=[
                "*",
                ("flap_defl", "flap_defl_up"),
                ("slat_defl", "slat_defl_up"),
            ],
            promotes_outputs=[("CL_max", Mission.Design.LIFT_COEFFICIENT_MAX_FLAPS_UP)],
        )
        self.add_subsystem(
            "flaps_takeoff",
            FlapsGroup(aviary_options=aviary_options),
            # slat deflection same for takeoff and landing
            promotes_inputs=["*", ("flap_defl", Aircraft.Wing.FLAP_DEFLECTION_TAKEOFF),
                             ("slat_defl", Aircraft.Wing.MAX_SLAT_DEFLECTION_TAKEOFF)],
            promotes_outputs=[
                ("CL_max", Mission.Takeoff.LIFT_COEFFICIENT_MAX),
                (
                    "delta_CL",
                    Mission.Takeoff.LIFT_COEFFICIENT_FLAP_INCREMENT,
                ),
                (
                    "delta_CD",
                    Mission.Takeoff.DRAG_COEFFICIENT_FLAP_INCREMENT,
                ),
            ],
        )
        self.add_subsystem(
            "flaps_landing",
            FlapsGroup(aviary_options=aviary_options),
            promotes_inputs=["*", ("flap_defl", Aircraft.Wing.FLAP_DEFLECTION_LANDING),
                             ("slat_defl", Aircraft.Wing.MAX_SLAT_DEFLECTION_LANDING)],
            promotes_outputs=[
                ("CL_max", Mission.Landing.LIFT_COEFFICIENT_MAX),
                (
                    "delta_CL",
                    Mission.Landing.LIFT_COEFFICIENT_FLAP_INCREMENT,
                ),
                (
                    "delta_CD",
                    Mission.Landing.DRAG_COEFFICIENT_FLAP_INCREMENT,
                ),
            ],
        )

        self.set_input_defaults("alt_flaps", 0)
        self.set_input_defaults("flap_defl_up", 0)
        self.set_input_defaults("slat_defl_up", 0)



import numpy as np
import openmdao.api as om
from openmdao.components.interp_util.interp import InterpND

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class DetailedWingBendingFact(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        aviary_options: AviaryValues = self.options['aviary_options']
        input_station_dist = aviary_options.get_val(Aircraft.Wing.INPUT_STATION_DIST)
        num_input_stations = len(input_station_dist)
        num_wing_engines = aviary_options.get_val(Aircraft.Engine.NUM_WING_ENGINES)
        count = len(aviary_options.get_val('engine_models'))

        add_aviary_input(self, Aircraft.Wing.LOAD_PATH_SWEEP_DIST,
                         val=np.zeros(num_input_stations - 1))

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_DIST,
                         val=np.zeros(num_input_stations))

        add_aviary_input(self, Aircraft.Wing.CHORD_PER_SEMISPAN_DIST,
                         val=np.zeros(num_input_stations))

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Engine.POD_MASS, val=0.0)

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=0.0)

        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO_REF, val=0.0)

        add_aviary_input(self, Aircraft.Wing.STRUT_BRACING_FACTOR, val=0.0)

        add_aviary_input(self, Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR, val=0.0)

        add_aviary_input(self, Aircraft.Engine.WING_LOCATIONS,
                         val=np.zeros([count, int(num_wing_engines/2)]))

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, val=0.0)

        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_REF, val=0.0)

        add_aviary_output(self, Aircraft.Wing.BENDING_FACTOR, val=0.0)

        add_aviary_output(self, Aircraft.Wing.ENG_POD_INERTIA_FACTOR, val=0.0)

    def setup_partials(self):
        # TODO: Analytic derivs will be challenging, but possible.
        self.declare_partials("*", "*", method='cs')

    def compute(self, inputs, outputs):
        aviary_options: AviaryValues = self.options['aviary_options']
        input_station_dist = aviary_options.get_val(Aircraft.Wing.INPUT_STATION_DIST)
        inp_stations = np.array(input_station_dist)
        num_integration_stations = \
            aviary_options.get_val(Aircraft.Wing.NUM_INTEGRATION_STATIONS)
        num_wing_engines = aviary_options.get_val(Aircraft.Engine.NUM_WING_ENGINES)

        # TODO: Support all options for this parameter.
        # 0.0 : input distribution
        # 1.0 : triangular distribution
        # 2.0 : elliptical distribution (default)
        # 3.0 : rectangular distribution
        # 1.0-2.0 : blend of triangular and elliptical
        # 2.0-3.0 : blend of elliptical and rectangular
        load_distribution_factor = \
            aviary_options.get_val(Aircraft.Wing.LOAD_DISTRIBUTION_CONTROL)

        load_path_sweep = inputs[Aircraft.Wing.LOAD_PATH_SWEEP_DIST]
        thickness_to_chord = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_DIST]
        chord = inputs[Aircraft.Wing.CHORD_PER_SEMISPAN_DIST]
        engine_locations = inputs[Aircraft.Engine.WING_LOCATIONS]
        gross_mass = inputs[Mission.Design.GROSS_MASS]
        # TODO currently using pod mass of engines even if not mounted on wing
        pod_mass = inputs[Aircraft.Engine.POD_MASS]
        fstrt = inputs[Aircraft.Wing.STRUT_BRACING_FACTOR]
        faert = inputs[Aircraft.Wing.AEROELASTIC_TAILORING_FACTOR]

        ar = inputs[Aircraft.Wing.ASPECT_RATIO]
        arref = inputs[Aircraft.Wing.ASPECT_RATIO_REF]
        tc = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        tcref = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_REF]

        # NOTE look at leaps1 code for examples on multi-engine support
        #    - will require list of pod masses for each engine
        # Currently implementation only partially addresses issues - odd numbers of wing
        # mounted engines pretend the "odd" engine out is not on the wing and is ignored
        # There are also no checks that number of engine locations is consistent with
        # half of number of wing mounted engines, which should get added to preprocessor

        # flatten engine variables into single lists, sorty by wing location
        engine_locations = engine_locations.flatten()
        # there should be a pod mass for every engine location
        pod_mass = np.repeat(pod_mass.flatten(), (1+num_wing_engines) % 2)
        engine_data = np.vstack((pod_mass, engine_locations))
        engine_data = engine_data.transpose()[np.lexsort(engine_data)]
        pod_mass = engine_data[:, 0]
        engine_locations = engine_data[:, 1]

        target_dy = (inp_stations[-1] - inp_stations[0]) / num_integration_stations
        stations_per_section = np.floor(np.abs(np.diff(inp_stations) / target_dy + 0.5))
        stations_per_section[-1] += 1  # add one more point to the last section
        integration_stations = np.empty(0, dtype=chord.dtype)
        sweep_int_stations = np.empty(0, dtype=chord.dtype)

        for i, val in enumerate(inp_stations[1:]):
            endpoint = i == len(inp_stations) - 2
            per_section = int(stations_per_section[i])
            integration_stations = np.append(
                integration_stations,
                np.linspace(inp_stations[i], val, per_section, endpoint=endpoint))
            sweep_int_stations = np.append(sweep_int_stations,
                                           load_path_sweep[i] * np.ones(per_section))

        dy = np.diff(integration_stations)
        avg_sweep = np.sum((dy[1:] + 2.0 * integration_stations[1:-1]) * dy[1:] *
                           sweep_int_stations[1:-1])

        # TODO: add all load_distribution_factor options
        if load_distribution_factor == 1:
            load_intensity = 1.0 - integration_stations
        elif load_distribution_factor == 2:
            load_intensity = np.sqrt(1.0 - integration_stations ** 2)
        elif load_distribution_factor == 3:
            load_intensity = np.ones(num_integration_stations + 1)
        else:
            raise om.AnalysisError(
                f'{load_distribution_factor} is not a valid value for {Aircraft.Wing.LOAD_DISTRIBUTION_CONTROL}, it must be "1", "2", or "3".')

        chord_interp = InterpND(method='slinear', points=(inp_stations),
                                x_interp=integration_stations)
        chord_int_stations = chord_interp.evaluate_spline(
            chord, compute_derivative=False)
        if arref > 0.0:
            # Scale
            chord_int_stations *= arref / ar

        del_load = dy * (
            chord_int_stations[:-1] * (2*load_intensity[:-1] + load_intensity[1:]) +
            chord_int_stations[1:] * (2*load_intensity[1:] + load_intensity[:-1])) / 6

        el = np.sum(del_load)

        del_moment = dy**2 * (
            chord_int_stations[:-1] * (load_intensity[:-1]+load_intensity[1:]) +
            chord_int_stations[1:] * (3*load_intensity[1:]+load_intensity[:-1])) / 12

        load_path_length = np.flip(
            np.append(np.zeros(1), np.cumsum(np.flip(del_load)[:-1])))
        csw = 1. / np.cos(sweep_int_stations[:-1] * np.pi/180.)
        emi = (del_moment + dy * load_path_length) * csw
        # em = np.sum(emi)

        tc_interp = InterpND(method='slinear', points=(inp_stations),
                             x_interp=integration_stations)
        tc_int_stations = tc_interp.evaluate_spline(
            thickness_to_chord, compute_derivative=False)
        if tcref > 0.0:
            tc_int_stations *= tc / tcref

        total_moment = np.cumsum(emi[::-1])[::-1]

        bma = total_moment * csw / (chord_int_stations[:-1] * tc_int_stations[:-1])

        pm = np.sum((bma[:-1] + bma[1:]) * dy[:-1] * 0.5)

        # s = np.sum((chord_int_stations[:-1] + chord_int_stations[1:]) * dy * 0.5)

        btb = 4 * pm / el

        sa = np.sin(avg_sweep * np.pi / 180.)
        if ar <= 5.0:
            caya = 0.0
        else:
            caya = ar - 5.0

        bt = btb / (ar**(0.25*fstrt) * (1.0 + (0.5*faert - 0.16*fstrt)
                    * sa**2 + 0.03*caya * (1.0-0.5*faert)*sa))
        outputs[Aircraft.Wing.BENDING_FACTOR] = bt

        eel = np.zeros(len(dy) + 1)
        loc = np.where(integration_stations < engine_locations[0])[0]
        eel[loc] = 1.0

        delme = dy * eel[1:]
        delme[loc[-1]] = engine_locations[0] - integration_stations[loc[-1]]

        eem = delme * csw
        eem = np.cumsum(eem[::-1])[::-1]

        ea = eem * csw / (chord_int_stations[:-1] * tc_int_stations[:-1])

        bte = 8 * np.sum((ea[:-1] + ea[1:]) * dy[:-1] * 0.5)

        outputs[Aircraft.Wing.ENG_POD_INERTIA_FACTOR] = 1.0 - \
            bte / bt * pod_mass / gross_mass



"""
OpenMDAO system for generating the aero tables that were typically printed in FLOPS.
"""
import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Mission


# TODO : This will need to create an instance of the dynamic aero group, perhaps as a
# subproblem, and run it at the table of mach numbers and lift coefficients. Right now,
# it is a placeholder, and also serves as a sink for all parts of the aircraft data
# structures that are passed to the dynamic portion, so that they can be overridden if
# needed.
class AeroReport(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(self, Aircraft.Canard.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Nacelle.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.WETTED_AREA, 0.0)
        add_aviary_input(self, Aircraft.Wing.WETTED_AREA, 0.0)

        add_aviary_input(self, Aircraft.Canard.FINENESS, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.FINENESS, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.FINENESS, 0.0)
        add_aviary_input(self, Aircraft.Nacelle.FINENESS, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.FINENESS, 0.0)
        add_aviary_input(self, Aircraft.Wing.FINENESS, 0.0)

        add_aviary_input(self, Aircraft.Canard.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_input(self, Aircraft.Fuselage.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_input(self, Aircraft.HorizontalTail.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_input(self, Aircraft.Nacelle.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_input(self, Aircraft.VerticalTail.CHARACTERISTIC_LENGTH, 0.0)
        add_aviary_input(self, Aircraft.Wing.CHARACTERISTIC_LENGTH, 0.0)

        add_aviary_input(self, Mission.Design.MACH, 0.0)
        add_aviary_input(self, Mission.Design.LIFT_COEFFICIENT, 0.0)



'''
Define utilities for calculating takeoff aerodynamics.
'''
from collections.abc import Sequence

import numpy as np
import openmdao.api as om
import scipy.constants as _units

from aviary.subsystems.aerodynamics.flops_based.ground_effect import \
    GroundEffect
from aviary.subsystems.aerodynamics.gasp_based.common import AeroForces
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class TakeoffAeroGroup(om.Group):
    '''
    Define a group for calculating takeoff aerodynamics.
    '''

    def initialize(self):
        options = self.options

        options.declare('num_nodes', default=1, types=int, lower=0)

        options.declare(
            'ground_altitude', default=0., types=float,
            desc='true altitude of the ground from mean sea level (m)'
        )

        options.declare(
            'use_spoilers', default=False, types=bool,
            desc='true for spoilers deployed'
        )

        options.declare(
            'spoiler_drag_coefficient', default=0.0, types=float,
            desc='spoiler drag coefficitnt'
        )

        options.declare(
            'spoiler_lift_coefficient', default=0.0, types=float,
            desc='spoiler lift coefficitnt'
        )

        options.declare(
            'angles_of_attack', types=Sequence,
            desc='sequence of angles of attack (deg); at least two values required'
        )

        options.declare(
            'lift_coefficients', types=Sequence,
            desc='sequence of lift coefficients, one for each angle of attack'
        )

        options.declare(
            'drag_coefficients', types=Sequence,
            desc='sequence of drag coefficients, one for each angle of attack'
        )

        # NOTE: FLOPS did not enforce a lower bound.
        # - Should Aviary enforce a lower bound?
        options.declare(
            'lift_coefficient_factor', default=1., types=float,
            desc='factor for takeoff lift'
        )

        options.declare(
            'drag_coefficient_factor', default=1., types=float,
            desc='factor for takeoff drag'
        )

        options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        options = self.options

        aviary_options = options['aviary_options']

        nn = options['num_nodes']
        angles_of_attack = np.array(options['angles_of_attack']) * _units.degree

        lift_coefficient_factor = options['lift_coefficient_factor']

        lift_coefficients = \
            np.array(options['lift_coefficients']) * lift_coefficient_factor

        drag_coefficient_factor = options['drag_coefficient_factor']

        drag_coefficients = \
            np.array(options['drag_coefficients']) * drag_coefficient_factor

        type_ = om.MetaModelSemiStructuredComp
        inputs = ['angle_of_attack']

        takeoff_polar: type_ = self.add_subsystem(
            'takeoff_polar',
            type_(method='slinear', extrapolate=False, vec_size=nn),
            promotes_inputs=inputs
        )

        takeoff_polar.add_input('angle_of_attack', angles_of_attack, units='rad')

        takeoff_polar.add_output('lift_coefficient', lift_coefficients, units='unitless')
        takeoff_polar.add_output('drag_coefficient', drag_coefficients, units='unitless')

        ground_altitude = options['ground_altitude']

        kwargs = {
            'num_nodes': nn,
            'ground_altitude': ground_altitude,
        }

        inputs = [
            'angle_of_attack', Dynamic.Mission.ALTITUDE, Dynamic.Mission.FLIGHT_PATH_ANGLE,
            ('minimum_drag_coefficient', Mission.Takeoff.DRAG_COEFFICIENT_MIN),
            Aircraft.Wing.ASPECT_RATIO, Aircraft.Wing.HEIGHT,
            Aircraft.Wing.SPAN
        ]

        self.add_subsystem(
            'ground_effect', GroundEffect(**kwargs), promotes_inputs=inputs
        )

        self.connect(
            'takeoff_polar.lift_coefficient', 'ground_effect.base_lift_coefficient'
        )

        self.connect(
            'takeoff_polar.drag_coefficient', 'ground_effect.base_drag_coefficient'
        )

        gear_drag = aviary_options.get_val(Aircraft.LandingGear.DRAG_COEFFICIENT)

        f = f'climb_drag_coefficient = ground_effect_drag + {gear_drag}'

        if options['use_spoilers']:
            spoiler_drag = options['spoiler_drag_coefficient']
            spoiler_lift = options['spoiler_lift_coefficient']
            f = f + f' + {spoiler_drag}'

            self.add_subsystem(
                'add_extra_lift_coefficients',
                om.ExecComp(
                    f'climb_lift_coefficient = ground_effect_lift + {spoiler_lift}',
                    climb_lift_coefficient={'units': 'unitless', 'shape': nn},
                    ground_effect_lift={'units': 'unitless', 'shape': nn}),
                promotes_inputs=['ground_effect_lift'],
                promotes_outputs=['climb_lift_coefficient'])

            self.connect('ground_effect.lift_coefficient', 'ground_effect_lift')
            self.connect('climb_lift_coefficient', 'aero_forces.CL')

        else:
            self.connect('ground_effect.lift_coefficient', 'aero_forces.CL')

        self.add_subsystem(
            'add_extra_drag_coefficients',
            om.ExecComp(
                f,
                climb_drag_coefficient={'units': 'unitless', 'shape': nn},
                ground_effect_drag={'units': 'unitless', 'shape': nn}),
            promotes_inputs=['ground_effect_drag'],
            promotes_outputs=['climb_drag_coefficient'])

        self.connect('ground_effect.drag_coefficient', 'ground_effect_drag')
        self.connect('climb_drag_coefficient', 'aero_forces.CD')

        inputs = [Dynamic.Mission.DYNAMIC_PRESSURE, Aircraft.Wing.AREA]
        outputs = [Dynamic.Mission.LIFT, Dynamic.Mission.DRAG]

        self.add_subsystem(
            'aero_forces', AeroForces(num_nodes=nn),
            promotes_inputs=inputs, promotes_outputs=outputs
        )

        self.set_input_defaults('angle_of_attack', np.zeros(nn), 'rad')



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class EngineMiscMass(om.ExplicitComponent):
    '''
    Calculates miscellaneous mass for a set of a engine model.

    Assumptions
    -----------
    Calculates total sum of all misc engine mass on the aircraft

    Currently using engine-level additional mass (scaled by num_engines)
    and propulsion-level starter and controls mass, not multi-engine safe!
    '''

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options')

    def setup(self):
        add_aviary_input(
            self, Aircraft.Engine.ADDITIONAL_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.MISC_MASS_SCALER, val=0.0)
        add_aviary_input(
            self, Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS, val=0.0)
        add_aviary_input(self, Aircraft.Propulsion.TOTAL_STARTER_MASS, val=0.0)

        add_aviary_output(self, Aircraft.Propulsion.TOTAL_MISC_MASS, val=0.0)

        self.declare_partials(
            of=Aircraft.Propulsion.TOTAL_MISC_MASS,
            wrt=[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS,
                 Aircraft.Propulsion.TOTAL_STARTER_MASS,
                 Aircraft.Engine.ADDITIONAL_MASS,
                 Aircraft.Propulsion.MISC_MASS_SCALER],
            val=1
        )

    def compute(self, inputs, outputs):
        # TODO temporarily using engine-level additional mass and multiplying
        #      by num_engines to get propulsion-level additional mass
        options: AviaryValues = self.options['aviary_options']
        num_engines = options.get_val(Aircraft.Engine.NUM_ENGINES)

        addtl_mass = sum(inputs[Aircraft.Engine.ADDITIONAL_MASS] * num_engines)
        ctrl_mass = inputs[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS]
        starter_mass = inputs[Aircraft.Propulsion.TOTAL_STARTER_MASS]
        scaler = inputs[Aircraft.Propulsion.MISC_MASS_SCALER]

        misc_mass = (starter_mass + addtl_mass + ctrl_mass) * scaler

        outputs[Aircraft.Propulsion.TOTAL_MISC_MASS] = misc_mass

    def compute_partials(self, inputs, J):
        # TODO temporarily using engine-level additional mass and multiplying
        #      by num_engines to get propulsion-level additional mass
        options: AviaryValues = self.options['aviary_options']
        num_engines = options.get_val(Aircraft.Engine.NUM_ENGINES)

        addtl_mass = inputs[Aircraft.Engine.ADDITIONAL_MASS] * num_engines
        ctrl_mass = inputs[Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS]
        starter_mass = inputs[Aircraft.Propulsion.TOTAL_STARTER_MASS]
        scaler = inputs[Aircraft.Propulsion.MISC_MASS_SCALER]

        J[Aircraft.Propulsion.TOTAL_MISC_MASS,
          Aircraft.Propulsion.TOTAL_ENGINE_CONTROLS_MASS
          ] = scaler

        J[Aircraft.Propulsion.TOTAL_MISC_MASS,
          Aircraft.Propulsion.TOTAL_STARTER_MASS
          ] = scaler

        J[Aircraft.Propulsion.TOTAL_MISC_MASS,
          Aircraft.Engine.ADDITIONAL_MASS
          ] = scaler * num_engines

        J[Aircraft.Propulsion.TOTAL_MISC_MASS,
          Aircraft.Propulsion.MISC_MASS_SCALER
          ] = starter_mass + addtl_mass + ctrl_mass



import unittest

import openmdao.api as om
from parameterized import parameterized

from aviary.subsystems.mass.flops_based.passenger_service import (
    AltPassengerServiceMass, PassengerServiceMass)
from aviary.utils.test_utils.variable_test import assert_match_varnames
from aviary.validation_cases.validation_tests import (Version,
                                                      flops_validation_test,
                                                      get_flops_case_names,
                                                      get_flops_inputs,
                                                      print_case)
from aviary.variable_info.variables import Aircraft, Mission


class PassengerServiceMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'passenger_service_weight',
            PassengerServiceMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=[Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER,
                        Mission.Design.RANGE],
            output_keys=Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            version=Version.TRANSPORT,
            tol=2e-4)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


class AlternatePassengerServiceMassTest(unittest.TestCase):

    def setUp(self):
        self.prob = om.Problem()

    @parameterized.expand(get_flops_case_names(),
                          name_func=print_case)
    def test_case(self, case_name):

        prob = self.prob

        prob.model.add_subsystem(
            'alternate_passenger_service_weight',
            AltPassengerServiceMass(aviary_options=get_flops_inputs(case_name)),
            promotes_inputs=['*'],
            promotes_outputs=['*'],
        )

        prob.setup(check=False, force_alloc_complex=True)

        flops_validation_test(
            prob,
            case_name,
            input_keys=Aircraft.CrewPayload.PASSENGER_SERVICE_MASS_SCALER,
            output_keys=Aircraft.CrewPayload.PASSENGER_SERVICE_MASS,
            version=Version.ALTERNATE)

    def test_IO(self):
        assert_match_varnames(self.prob.model)


if __name__ == "__main__":
    unittest.main()



import numpy as np


def almost_equal(a, b, rel_tol=1e-8, abs_tol=0.0):
    if isinstance(a, (float, np.float32, np.float64)) and isinstance(b, (float, np.float32, np.float64)):
        return np.isclose(a, b, rtol=rel_tol, atol=abs_tol)

    if isinstance(a, np.ndarray) and isinstance(b, np.ndarray):
        return np.allclose(a, b, rtol=rel_tol, atol=abs_tol)

    if isinstance(a, dict) and isinstance(b, dict):
        if a.keys() != b.keys():
            return False
        for key in a.keys():
            if not almost_equal(a[key], b[key]):
                return False
        return True

    return a == b


def merge_2_meta_data_dicts(dict1, dict2):
    '''
    Combines metadata from two dictionaries into a single dictionary containing metadata of both.

    Performs a check to ensure that the two dictionaries don't have the same variable with conflicting
    metadata. Assuming that check passes, the variables and their associated metadata from both
    dictionaries are combined into one dictionary.

    Parameters
    ----------
    dict1, dict2 : dict
        Dictionaries with keys of Aviary variables and values of their associated metadata.

    Returns
    -------
    new_dict : dict
        Single dictionary with all the combined entries from the two input Aviary metadata dictionaries.

    Raises
    ------
    ValueError
        Raises error if dict1 and dict2 have differing metadata for the same variable.
    '''

    dict1_keys = set(dict1.keys())  # get keys of provided dictionaries
    dict2_keys = set(dict2.keys())  # get keys of provided dictionaries
    # get keys that are unique to each dictionary
    all_unique_keys = (dict1_keys | dict2_keys) - (dict1_keys & dict2_keys)
    # get all keys that are repeated in both dictionaries
    all_duplicate_keys = dict1_keys & dict2_keys
    # merge the two dictionaries, and allow the repeated keys to be overwritten
    merged = {**dict1, **dict2}

    dict_of_unique_variables = dict(
        # get a dictionary of all the unique keys (and their values) from the input dictionaries
        map(lambda key: (key, merged.get(key, None)), all_unique_keys))

    # initialize a dictionary of the values with duplicated keys in both input dictionaries
    dict_of_duplicate_variables = {}

    for key in sorted(all_duplicate_keys):
        # check that the metadata in both dictionaries associated with the same key is the same
        value1, value2 = dict1[key], dict2[key]
        # throw an error if the dicts have the same key with different metadata
        if not almost_equal(value1, value2):
            raise ValueError(
                f'You have attempted to merge metadata dictionaries that contain the same variable with different metadata. The offending variable present in multiple dictionaries is "{key}".')

        # add the key and metadata to the dictionary of duplicate variables
        dict_of_duplicate_variables[key] = dict1[key]

    # merge together the final dictionary
    new_dict = {**dict_of_duplicate_variables, **dict_of_unique_variables}

    return new_dict


def merge_meta_data(dicts_to_merge):
    '''
    Merges the metadata of multiple Aviary metadata dictionaries into a single metadata dictionary.

    Checks the metadata of all the provided Aviary metadata dictionaries to see if there are identical
    variables with conflicting metadata. Assuming this check passes, the input dictionaries are merged
    together into a single dictionary containing all their information.

    Parameters
    ----------
    dicts_to_merge : list of dicts
        List of Aviary metadata dictionaries to be merged.

    Returns
    -------
    merged_dict : dict
        Single Aviary metadata dictionary with all the information of the inputted metadata dictionaries.

    Raises
    ------
    None
        No exceptions raised by this method, although other methods called within may raise exceptions.
    '''
    merged_dict = dicts_to_merge[0]

    # iterate through all the dictionaries and merge them together into one, two by two
    for dict_item in dicts_to_merge:
        merged_dict = merge_2_meta_data_dicts(merged_dict, dict_item)

    return merged_dict



import os

import numpy as np
import openmdao.api as om

from pathlib import Path

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.aerodynamics.gasp_based.common import AeroForces, TimeRamp
from aviary.utils.named_values import NamedValues, get_keys
from aviary.utils.data_interpolator_builder import build_data_interpolator
from aviary.utils.csv_data_file import read_data_file
from aviary.utils.functions import get_path
from aviary.utils.named_values import get_items
from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic, Mission

# Map of variable names to allowed headers for data files (only lowercase required,
# spaces are replaced with underscores when data tables are read)
# "Repeated" aliases allows variables with different cases to match with desired
# all-lowercase name
aliases = {Dynamic.Mission.ALTITUDE: ['h', 'alt', 'altitude'],
           Dynamic.Mission.MACH: ['m', 'mach'],
           'angle_of_attack': ['alpha', 'angle_of_attack', 'AoA'],
           'flap_deflection': ['flap_deflection'],
           'hob': ['hob'],
           'lift_coefficient': ['cl', 'lift_coefficient'],
           'drag_coefficient': ['cd', 'drag_coefficient'],
           'delta_lift_coefficient': ['delta_cl', 'dcl'],
           'delta_drag_coefficient': ['delta_cd', 'dcd']
           }


class CruiseAero(om.Group):
    """Free-air lift and drag using a table lookup."""

    def initialize(self):
        self.options.declare('num_nodes', default=1, types=int)

        self.options.declare('aero_data', types=(str, Path, NamedValues), default=None,
                             desc='Data file or NamedValues object containing lift and '
                                  'drag coefficient table as a function of altitude, '
                                  'Mach, and angle of attack')

        self.options.declare('training_data', default=False,
                             desc='When True, the aero tables will be passed as '
                                  'OpenMDAO variables')

        self.options.declare('structured', types=bool, default=True,
                             desc='Flag that sets if data is a structured grid')

        self.options.declare('extrapolate', default=True, desc='Flag that sets if drag '
                                                               'data can be extrapolated')

    def setup(self):
        options = self.options
        nn = options['num_nodes']
        aero_data = options['aero_data']
        training_data = options['training_data']
        structured = options['structured']
        extrapolate = options['extrapolate']

        # handle aliasing for training data
        extra_promotes = []
        if training_data:
            extra_promotes = [('lift_coefficient_train', Aircraft.Design.LIFT_POLAR),
                              ('drag_coefficient_train', Aircraft.Design.DRAG_POLAR)]

        interp_comp = _build_free_aero_interp(num_nodes=nn, aero_data=aero_data,
                                              training_data=training_data,
                                              structured=structured,
                                              extrapolate=extrapolate)

        self.add_subsystem('free_aero_interp',
                           subsys=interp_comp,
                           promotes_inputs=[Dynamic.Mission.ALTITUDE,
                                            Dynamic.Mission.MACH,
                                            ('angle_of_attack', 'alpha')]
                           + extra_promotes,
                           promotes_outputs=[('lift_coefficient', 'CL'), ('drag_coefficient', 'CD')])

        self.add_subsystem("forces", AeroForces(num_nodes=nn), promotes=["*"])

        self.set_input_defaults(Dynamic.Mission.MACH, np.zeros(nn))


class LowSpeedAero(om.Group):
    """Lift and drag near the ground using a table lookup.

    Includes increments due to ground effects, landing gear, and flaps. Retraction or
    application of landing gear and flaps to vary increments over time are controlled by
    options.
    """

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

        self.options.declare('free_aero_data', types=(str, Path, NamedValues),
                             default=None,
                             desc='Data file or NamedValues object containing free aero '
                                  'lift and drag coefficient table as a function of '
                                  'altitude, Mach, and angle of attack')
        self.options.declare('flaps_aero_data', types=(str, Path, NamedValues),
                             default=None,
                             desc='Data file or NamedValues object containing flaps aero '
                                  'lift and drag coefficient table as a function of '
                                  'altitude, Mach, and angle of attack')
        self.options.declare('ground_aero_data', types=(str, Path, NamedValues),
                             default=None,
                             desc='Data file or NamedValues object containing ground '
                                  'aero lift and drag coefficient table as a function '
                                  'of altitude, Mach, and angle of attack')

        self.options.declare('training_data', types=bool, default=False,
                             desc='When True, all aero tables will be passed as '
                                  'OpenMDAO variables')
        self.options.declare('structured', types=bool, default=True,
                             desc='Flag that sets if all data are structured grids')
        self.options.declare('extrapolate', types=bool, default=False,
                             desc='Flag that sets if all drag data can be extrapolated')

        self.options.declare("retract_gear", default=True, types=bool,
                             desc="True to start with gear landing gear down, False "
                                  "for reverse")
        self.options.declare("retract_flaps", default=True, types=bool,
                             desc="True to start with flaps applied, False for reverse")

    def setup(self):
        options = self.options
        nn = options['num_nodes']
        free_aero_data = options['free_aero_data']
        flaps_aero_data = options['flaps_aero_data']
        ground_aero_data = options['ground_aero_data']
        training_data = options['training_data']
        structured = options['structured']
        extrapolate = options['extrapolate']

        # convert altitude to height/span for ground effects
        hob = om.ExecComp(
            "hob = (wing_height + altitude - airport_alt) / wingspan",
            altitude=dict(val=np.zeros(nn), units="ft"),
            airport_alt=dict(val=0, units="ft"),
            wingspan=dict(val=117.8, units="ft"),
            wing_height=dict(
                val=8.0, units="ft", desc="Wing height above ground at 0 alt"
            ),
            hob=dict(shape=(nn,), desc="Wing height/span", units="unitless"),
        )
        self.add_subsystem(
            "hob",
            hob,
            promotes_inputs=[
                Dynamic.Mission.ALTITUDE,
                "airport_alt",
                ("wingspan", Aircraft.Wing.SPAN),
                ("wing_height", Aircraft.Wing.HEIGHT),
            ],
            promotes_outputs=["hob"],
        )

        free_aero_interp = _build_free_aero_interp(nn, aero_data=free_aero_data,
                                                   training_data=training_data,
                                                   structured=structured,
                                                   extrapolate=extrapolate)

        # "base" free-air coefficients
        self.add_subsystem(
            "interp_free",
            free_aero_interp,
            promotes_inputs=[Dynamic.Mission.ALTITUDE,
                             Dynamic.Mission.MACH, ('angle_of_attack', 'alpha')],
            promotes_outputs=[
                ("lift_coefficient", "CL_free"),
                ("drag_coefficient", "CD_free"),
                ("lift_coefficient_max", "CL_max_free"),
            ],
        )

        flaps_aero_interp = _build_flaps_aero_interp(nn, aero_data=flaps_aero_data,
                                                     training_data=training_data,
                                                     structured=structured,
                                                     extrapolate=extrapolate)

        # flap drag and lift increment from full flap deflection
        self.add_subsystem(
            "interp_flaps",
            flaps_aero_interp,
            promotes_inputs=[('flap_deflection', 'flap_defl'),
                             Dynamic.Mission.MACH, ('angle_of_attack', 'alpha')],
            promotes_outputs=[
                ("delta_lift_coefficient", "dCL_flaps_full"),
                ("delta_drag_coefficient", "dCD_flaps_full"),
                ("delta_lift_coefficient_max", "dCL_max_flaps"),
            ],
        )

        ground_aero_interp = _build_ground_aero_interp(nn, aero_data=ground_aero_data,
                                                       training_data=training_data,
                                                       structured=structured,
                                                       extrapolate=extrapolate)

        # drag and lift increments from ground effects
        self.add_subsystem(
            "interp_ground",
            ground_aero_interp,
            promotes_inputs=[Dynamic.Mission.MACH, ('angle_of_attack', 'alpha'), 'hob'],
            promotes_outputs=[
                ("delta_lift_coefficient", "dCL_ground"),
                ("delta_drag_coefficient", "dCD_ground"),
                ("delta_lift_coefficient_max", "dCL_max_ground"),
            ],
        )

        # drag increment from landing gear
        self.add_subsystem(
            "gear_drag",
            GearDragIncrement(num_nodes=nn),
            promotes_inputs=["aircraft:*", "flap_defl", "mission:*"],
            promotes_outputs=[("dCD", "dCD_gear_full")],
        )

        # scale gear drag increment over time
        self.add_subsystem(
            "gear_drag_ramp",
            TimeRamp(
                num_nodes=nn, num_inputs=1, ramp_up=not self.options["retract_gear"]
            ),
            promotes_inputs=[
                ("t_init", "t_init_gear"),
                ("duration", "dt_gear"),
                ("x", "dCD_gear_full"),
                "t_curr",
            ],
            promotes_outputs=[("y", "dCD_gear")],
        )

        # scale flap drag and lift
        # TODO mux these?
        self.add_subsystem(
            "flaps_drag_ramp",
            TimeRamp(
                num_nodes=nn, num_inputs=1, ramp_up=not self.options["retract_gear"]
            ),
            promotes_inputs=[
                ("t_init", "t_init_flaps"),
                ("duration", "dt_flaps"),
                ("x", "dCD_flaps_full"),
                "t_curr",
            ],
            promotes_outputs=[("y", "dCD_flaps")],
        )
        self.add_subsystem(
            "flaps_lift_ramp",
            TimeRamp(
                num_nodes=nn, num_inputs=1, ramp_up=not self.options["retract_gear"]
            ),
            promotes_inputs=[
                ("t_init", "t_init_flaps"),
                ("duration", "dt_flaps"),
                ("x", "dCL_flaps_full"),
                "t_curr",
            ],
            promotes_outputs=[("y", "dCL_flaps")],
        )

        # add up CL + increments, CD + increments, CL_max + increments
        adder = om.AddSubtractComp()
        adder.add_equation(
            "CL", vec_size=nn, units='unitless', input_names=["CL_free", "dCL_ground", "dCL_flaps"]
        )
        adder.add_equation(
            "CD",
            vec_size=nn,
            units='unitless',
            input_names=["CD_free", "dCD_ground", "dCD_flaps", "dCD_gear"],
        )
        adder.add_equation(
            "CL_max",
            vec_size=nn,
            units='unitless',
            input_names=["CL_max_free", "dCL_max_flaps", "dCL_max_ground"],
        )
        self.add_subsystem("coef_sum", adder, promotes=["*"])

        # convert coefficients to forces
        self.add_subsystem(
            "forces",
            AeroForces(num_nodes=nn),
            promotes_inputs=[
                "CL",
                "CD",
                Dynamic.Mission.DYNAMIC_PRESSURE,
            ]
            + ["aircraft:*"],
            promotes_outputs=[Dynamic.Mission.LIFT, Dynamic.Mission.DRAG],
        )

        if self.options["retract_gear"]:
            # takeoff defaults
            self.set_input_defaults("dt_gear", 7)
        else:
            # TODO default gear duration for landing?
            pass

        if self.options["retract_flaps"]:
            # takeoff defaults
            self.set_input_defaults("flap_defl", 10 * np.ones(nn))
            self.set_input_defaults("dt_flaps", 3)
        else:
            # landing defaults
            self.set_input_defaults("flap_defl", 40 * np.ones(nn))
            # TODO default flap duration for landing?

        self.set_input_defaults(Dynamic.Mission.ALTITUDE, np.zeros(nn))
        self.set_input_defaults(Dynamic.Mission.MACH, np.zeros(nn))


class GearDragIncrement(om.ExplicitComponent):
    """Gear drag coefficient increment.

    Constant for a given *full* flap deflection.
    """

    def initialize(self):
        self.options.declare("num_nodes", default=1, types=int)

    def setup(self):
        nn = self.options["num_nodes"]

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=175400.0)

        add_aviary_input(self, Aircraft.Wing.AREA, val=1370.3)

        # note flap_defl should be a constant (scalar) but interpolation breaks with
        # mixed scalar/vector inputs so you can just ``prob.set_val("flap_defl", 10)``
        self.add_input(
            "flap_defl",
            0.0,
            shape=nn,
            units="deg",
            desc="DFLAP: Flap deflection angle. This is the *full* flap "
            "deflection angle for a given flight segment (DFLPLD, "
            "DFLPTO).",
        )

        self.add_output(
            "dCD", units="unitless", shape=nn,
            desc="Drag coefficient increment for landing gear"
        )

    def setup_partials(self):
        nn = self.options["num_nodes"]
        arange = np.arange(nn)
        self.declare_partials("dCD", ["flap_defl"], rows=arange, cols=arange)
        self.declare_partials("dCD", [Mission.Design.GROSS_MASS, Aircraft.Wing.AREA])

    def compute(self, inputs, outputs):
        gross_mass_initial, wing_area, flap_defl = inputs.values()
        gross_wt_initial = gross_mass_initial * GRAV_ENGLISH_LBM

        # landing gear flat plate area
        grfe = 0.0033 * gross_wt_initial**0.785
        # landing gear CD increment for zero flap deflection
        grcd = grfe / wing_area

        outputs["dCD"] = grcd * (1 - 0.454545 * flap_defl / 50)

    def compute_partials(self, inputs, J):
        gross_mass_initial, wing_area, flap_defl = inputs.values()
        gross_wt_initial = gross_mass_initial * GRAV_ENGLISH_LBM
        grfe = 0.0033 * gross_wt_initial**0.785
        grcd = grfe / wing_area

        J["dCD", Mission.Design.GROSS_MASS] = (
            0.0033 * 0.785 * gross_wt_initial ** (0.785 - 1) / wing_area
        ) * (1 - 0.454545 * flap_defl / 50) * GRAV_ENGLISH_LBM
        J["dCD", Aircraft.Wing.AREA] = -(grfe / wing_area**2) * (
            1 - 0.454545 * flap_defl / 50
        )
        J["dCD", "flap_defl"] = -grcd * 0.454545 / 50


def _build_free_aero_interp(num_nodes=0, aero_data=None, training_data=False,
                            method='lagrange2', structured=True, extrapolate=True):
    """creates interpolation components for cruise aero"""
    # build_data_interpolator normally handles converting to filepath and reading
    # data, but here we need to query the data before building the component
    if isinstance(aero_data, str):
        aero_data = get_path(aero_data)
    if isinstance(aero_data, Path):
        aero_data = read_data_file(aero_data, aliases=aliases)

    # aero_data is modified in-place, deepcopy required
    interp_data = aero_data.deepcopy()

    interp_data = _structure_special_grid(interp_data)

    required_inputs = {Dynamic.Mission.ALTITUDE, Dynamic.Mission.MACH,
                       'angle_of_attack'}
    required_outputs = {'lift_coefficient', 'drag_coefficient'}

    missing_variables = []
    if not required_inputs <= get_keys(interp_data):
        missing_variables.append([key for key in
                                  required_inputs.difference(get_keys(interp_data))])
    if not training_data and not required_outputs <= get_keys(interp_data):
        missing_variables.append([key for key in
                                  required_outputs.difference(get_keys(interp_data))])
    if missing_variables:
        raise KeyError('GASP-based aerodynamics interpolation missing required '
                       f'variables: {missing_variables}')

    if training_data:
        method = 'lagrange2'
    else:
        method = '3D-lagrange2'

    # add the 3d metamodel to the group, promoting all variables
    interp_comp = build_data_interpolator(num_nodes=num_nodes,
                                          interpolator_data=interp_data,
                                          interpolator_outputs={'lift_coefficient': 'unitless',
                                                                'drag_coefficient': 'unitless'},
                                          method=method,
                                          structured=structured,
                                          training_data=training_data,
                                          extrapolate=extrapolate)

    if training_data:
        return interp_comp
    else:
        group = om.Group()
        group.add_subsystem('free_aero_interp', interp_comp, promotes=['*'])

        # approx CL max as that with highest alpha, assume alpha monotonically increasing
        # free aero CL at max alpha is the same across altitudes but ignore that for now
        cl_max = interp_data.get_val('lift_coefficient', 'unitless')[0, :, -1]
        # add a 1d metamodel for cl_max, promoting all variables
        meta_1d = om.MetaModelStructuredComp(method='1D-lagrange2',
                                             vec_size=num_nodes,
                                             extrapolate=extrapolate)
        meta_1d.add_input(Dynamic.Mission.MACH, 0.0, units="unitless",
                          shape=num_nodes,
                          training_data=interp_data.get_val(Dynamic.Mission.MACH,
                                                            'unitless'))
        meta_1d.add_output('lift_coefficient_max', units="unitless",
                           shape=num_nodes, training_data=cl_max)

        group.add_subsystem("lift_coefficient_max_interp", meta_1d, promotes=["*"])
        return group


def _build_flaps_aero_interp(num_nodes=0, aero_data=None, training_data=False,
                             method='slinear', structured=True, extrapolate=False):
    """creates interpolation components for cruise aero"""
    # TODO linear method default because standard GASP tables have only two flap
    #      deflections - may want to have option for two separate 2D tables instead?

    # build_data_interpolator normally handles converting to filepath and reading
    # data, but here we need to query the data before building the component
    if isinstance(aero_data, str):
        aero_data = get_path(aero_data)
    if isinstance(aero_data, Path):
        aero_data = read_data_file(aero_data, aliases=aliases)

    # aero_data is modified in-place, deepcopy required
    interp_data = aero_data.deepcopy()

    interp_data = _structure_special_grid(interp_data)

    required_inputs = {'flap_deflection', Dynamic.Mission.MACH, 'angle_of_attack'}
    required_outputs = {'delta_lift_coefficient', 'delta_drag_coefficient'}

    missing_variables = []
    if not required_inputs <= get_keys(interp_data):
        missing_variables.extend([key for key in
                                  required_inputs.difference(get_keys(interp_data))])
    if not training_data and not required_outputs <= get_keys(interp_data):
        missing_variables.extend([key for key in
                                  required_outputs.difference(get_keys(interp_data))])
    if missing_variables:
        raise KeyError('GASP-based aerodynamics interpolation missing required '
                       f'variables: {missing_variables}')

    dcl = interp_data.get_val('delta_lift_coefficient', 'unitless')
    defl = np.unique(interp_data.get_val('flap_deflection', 'deg')
                     )  # units don't matter, not using values
    alpha = np.unique(interp_data.get_val('angle_of_attack', 'deg')
                      )  # units don't matter, not using values
    mach = np.unique(interp_data.get_val(Dynamic.Mission.MACH, 'unitless'))

    dcl_max = np.zeros_like(dcl)
    shape = (defl.size, mach.size, alpha.size)
    dcl_max = np.resize(dcl_max, shape)
    dcl = np.reshape(dcl, shape)
    for i in range(defl.size):
        dcl_max[i, :, :] = np.broadcast_to(dcl[i, :, -1], (alpha.size, mach.size)).T

    interp_data.set_val('delta_lift_coefficient_max', dcl_max.flatten(), 'unitless')

    return build_data_interpolator(num_nodes=num_nodes,
                                   interpolator_data=interp_data,
                                   interpolator_outputs={'delta_lift_coefficient': 'unitless',
                                                         'delta_drag_coefficient': 'unitless',
                                                         'delta_lift_coefficient_max': 'unitless'},
                                   method=method,
                                   structured=structured,
                                   training_data=training_data,
                                   extrapolate=extrapolate)


def _build_ground_aero_interp(num_nodes=0, aero_data=None, training_data=False,
                              method='slinear', structured=True, extrapolate=True):
    """creates interpolation components for cruise aero"""
    # build_data_interpolator normally handles converting to filepath and reading
    # data, but here we need to query the data before building the component
    if isinstance(aero_data, str):
        aero_data = get_path(aero_data)
    if isinstance(aero_data, Path):
        aero_data = read_data_file(aero_data, aliases=aliases)

    # aero_data is modified in-place, deepcopy required
    interp_data = aero_data.deepcopy()

    required_inputs = {'hob', Dynamic.Mission.MACH, 'angle_of_attack'}
    required_outputs = {'delta_lift_coefficient', 'delta_drag_coefficient'}

    missing_variables = []
    if not required_inputs <= get_keys(interp_data):
        missing_variables.append([key for key in
                                  required_inputs.difference(get_keys(interp_data))])
    if not training_data and not required_outputs <= get_keys(interp_data):
        missing_variables.append([key for key in
                                  required_outputs.difference(get_keys(interp_data))])
    if missing_variables:
        raise KeyError('GASP-based aerodynamics interpolation missing required '
                       f'variables: {missing_variables}')

    dcl = interp_data.get_val('delta_lift_coefficient', 'unitless')
    alpha = np.unique(interp_data.get_val('angle_of_attack', 'deg')
                      )  # units don't matter, not using values
    mach = np.unique(interp_data.get_val(Dynamic.Mission.MACH, 'unitless'))
    hob = np.unique(interp_data.get_val('hob', 'unitless'))

    dcl_max = np.zeros_like(dcl)
    shape = (mach.size, hob.size, alpha.size)
    dcl_max = np.resize(dcl_max, shape)
    dcl = np.reshape(dcl, shape)
    for i in range(mach.size):
        dcl_max[i, :, :] = np.broadcast_to(dcl[i, :, -1], (alpha.size, hob.size)).T

    interp_data.set_val('delta_lift_coefficient_max', dcl_max.flatten(), 'unitless')

    # extrapolation fine especially for HOB over max
    return build_data_interpolator(num_nodes=num_nodes,
                                   interpolator_data=interp_data,
                                   interpolator_outputs={'delta_lift_coefficient': 'unitless',
                                                         'delta_drag_coefficient': 'unitless',
                                                         'delta_lift_coefficient_max': 'unitless'},
                                   method=method,
                                   structured=structured,
                                   training_data=training_data,
                                   extrapolate=extrapolate)


def _structure_special_grid(aero_data):
    """
    Structure a GASP-based data table that has a special case with incorrect number
    of alpha points for a structured grid

    This assumes that the first two values in the data table (or directly provided data)
    are the other independent variables (mach, alt, or hob depending on use case), while
    the third independent variable is alpha
    """
    # store these to later put data back in the NamedValues without knowing what's in it
    data = []
    keys = []
    units = []
    for idx, (key, (val, unit)) in enumerate(get_items(aero_data)):
        if idx < 2:
            data.append(val)
            keys.append(key)
            units.append(unit)
        else:
            break

    x0 = np.unique(data[0])
    x1 = np.unique(data[1])
    # units don't matter, not using values
    aoa = aero_data.get_val('angle_of_attack', 'deg')

    if data[0].shape[0] > x0.shape[0]:
        # special case alpha - this format saturates alpha at its max
        # get unique alphas at zero alt mach 0, should cover the full range
        mask = (data[0] == x0[0]) & (data[1] == x1[0])
        aoa = np.unique(aoa[mask])

    _, _, aoa = np.meshgrid(x0, x1, aoa)

    # put the aoa data back in the NamedValues object
    aero_data.set_val('angle_of_attack', aoa.flatten(), 'deg')

    return aero_data



import unittest

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials

from aviary.subsystems.aerodynamics.flops_based.skin_friction_drag import \
    SkinFrictionDrag
from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.variables import Aircraft


class SkinFrictionDragTest(unittest.TestCase):

    def test_derivs(self):
        nn = 2

        fine = np.array([0.13, .125, .1195, 10.0392, 1.5491, 1.5491])
        Re = np.array([[15.0, 11, 18.3, 183.4, 17.6, 17.6],
                      [17.0, 16, 15.3, 78.4, 23.6, 33.6]]) * 1e-6
        cf = np.array([[.00263, .00276, .00255, .00182, .00256, .00256],
                       [.00283, .00296, .00235, .00172, .00276, .00276]])
        Sw = np.array([2396.56, 592.65, 581.13, 4158.62, 273.45, 273.45])
        lam_up = np.array([6 * 0.03])
        lam_low = np.array([6 * 0.02])
        nc = len(fine)

        prob = om.Problem()
        model = prob.model

        options = AviaryValues()
        options.set_val(Aircraft.Fuselage.NUM_FUSELAGES, 1)
        options.set_val(Aircraft.Engine.NUM_ENGINES, [2])
        options.set_val(Aircraft.VerticalTail.NUM_TAILS, 1)
        options.set_val(Aircraft.Wing.AIRFOIL_TECHNOLOGY, 1.93)

        model.add_subsystem(
            'CDf', SkinFrictionDrag(num_nodes=nn, aviary_options=options),
            promotes_inputs=[Aircraft.Wing.AREA])

        prob.setup(force_alloc_complex=True)

        prob.set_val('CDf.fineness_ratios', fine)
        prob.set_val('CDf.Re', Re)
        prob.set_val('CDf.skin_friction_coeff', cf)
        prob.set_val('CDf.wetted_areas', Sw)
        prob.set_val('CDf.laminar_fractions_upper', lam_up)
        prob.set_val('CDf.laminar_fractions_lower', lam_low)
        prob.set_val(Aircraft.Wing.AREA, 198.0)

        derivs = prob.check_partials(out_stream=None, method="cs")

        # atol is low because Re magnitude is 1e8.
        # TODO: need to test outputs too
        assert_check_partials(derivs, atol=1e-7, rtol=1e-12)


if __name__ == "__main__":
    unittest.main()



# %%
import numpy as np
import openmdao.api as om
import scipy.constants as _units
from openmdao.components.interp_util.interp import InterpND

from aviary.variable_info.functions import add_aviary_input
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class BuffetLift(om.ExplicitComponent):
    """
    Computes lift margin before buffet onset.
    """

    def initialize(self):

        self.options.declare("num_nodes", default=1, types=int,
                             desc="Number of nodes along mission segment")

    def setup(self):
        nn = self.options["num_nodes"]

        # Simulation inputs
        self.add_input(
            Dynamic.Mission.MACH,
            shape=(nn),
            units='unitless',
            desc="Mach number")

        # Aero design inputs
        add_aviary_input(self, Mission.Design.MACH, 0.0)

        # Aircraft design inputs
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, 0.0)
        add_aviary_input(self, Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN, 0.0)
        add_aviary_input(self, Aircraft.Wing.SWEEP, 0.0)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD, 0.0)

        # Declare outputs
        self.add_output("DELCLB", shape=(nn), units='unitless',
                        desc="Delta lift coefficient before buffet onset")

    def setup_partials(self):
        nn = self.options["num_nodes"]

        self.declare_partials(
            'DELCLB',
            Dynamic.Mission.MACH,
            rows=np.arange(nn),
            cols=np.arange(nn))
        self.declare_partials('DELCLB', [Aircraft.Wing.ASPECT_RATIO,
                                         Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN,
                                         Aircraft.Wing.SWEEP,
                                         Aircraft.Wing.THICKNESS_TO_CHORD,
                                         Mission.Design.MACH])

    def compute(self, inputs, outputs):
        """
        Computes lift margin before buffet onset.
        """
        mach, design_Mach, AR, CAM, SW25, TC = inputs.values()

        del_Mach = mach - design_Mach

        TC23 = TC**(2.0/3.0)
        TC23 = np.repeat(TC23, len(del_Mach))
        x = np.transpose(np.array([TC23, del_Mach]))
        # dFCLB = [dFCLB_dTC23, dFCLB_dDELM]
        FCLB, dFCLB = BUFTTable.interpolate(x, compute_derivative=True)
        DELCLB = FCLB * (AR * (1.0 + CAM/10.0) / np.cos(SW25 / _units.degree))

        self.FCLB = FCLB
        self.dFCLB = dFCLB
        outputs['DELCLB'] = DELCLB

    def compute_partials(self, inputs, partials):

        TC = inputs[Aircraft.Wing.THICKNESS_TO_CHORD]
        SW25 = inputs[Aircraft.Wing.SWEEP]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        CAM = inputs[Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN]

        FCLB = self.FCLB
        dFCLB_dTC23 = self.dFCLB[:, 0]
        dFCLB_ddel_Mach = self.dFCLB[:, 1]

        a = _units.degree
        SW25_rad = SW25 / a
        cos_fact = 1.0 / np.cos(SW25_rad)

        dTC23_dTC = (2.0 / 3.0) * TC ** (-1.0 / 3.0)
        dDELCLB_dFCLB = AR * (1.0 + CAM/10.0) * cos_fact
        ddel_Mach_dMach = 1.0
        ddel_Mach_ddesign_Mach = -1.0

        # wrt Mach
        dCLB_dMach = dDELCLB_dFCLB * dFCLB_ddel_Mach * ddel_Mach_dMach

        # wrt design_Mach
        dCLB_ddesign_Mach = dDELCLB_dFCLB * dFCLB_ddel_Mach * ddel_Mach_ddesign_Mach

        # wrt AR
        dCLB_dAR = FCLB * (1.0 + CAM/10.0) * cos_fact

        # wrt TC
        dCLB_dTC = dDELCLB_dFCLB * dFCLB_dTC23 * dTC23_dTC

        # wrt SW25
        dCLB_dSW25 = (1.0 / a) * AR * FCLB * (1.0 + CAM/10.0) * \
            np.sin(SW25_rad) * cos_fact ** 2

        # wrt CAM
        dCLB_dCAM = FCLB * AR / 10.0 * cos_fact

        partials["DELCLB", Dynamic.Mission.MACH] = dCLB_dMach
        partials["DELCLB", Mission.Design.MACH] = dCLB_ddesign_Mach
        partials['DELCLB', Aircraft.Wing.ASPECT_RATIO] = dCLB_dAR
        partials['DELCLB', Aircraft.Wing.THICKNESS_TO_CHORD] = dCLB_dTC
        partials['DELCLB', Aircraft.Wing.SWEEP] = dCLB_dSW25
        partials['DELCLB', Aircraft.Wing.MAX_CAMBER_AT_70_SEMISPAN] = dCLB_dCAM

        return


BUFT = np.array(
    [[8010,  -0.8,  -0.6,  -0.4,  -0.3,  -0.2,  -0.1,   0.0,   0.05,    0.1,   0.15],
     [0.1, 0.078, 0.076, 0.061, 0.050, 0.046, 0.051, 0.070,   0.09,   0.12,  0.156],
     [0.12, 0.078, 0.076, 0.061, 0.050, 0.043, 0.036, 0.039,  0.051,  0.073,  0.104],
     [0.14, 0.078, 0.076, 0.061, 0.050, 0.042, 0.034, 0.030,  0.031,  0.040,  0.057],
     [0.16, 0.078, 0.076, 0.061, 0.050, 0.041, 0.031, 0.023,  0.020,  0.016,  0.010],
     [0.18, 0.078, 0.076, 0.061, 0.050, 0.040, 0.030, 0.018,  0.010,  0.002, -0.009],
     [0.20, 0.078, 0.076, 0.061, 0.050, 0.040, 0.029, 0.015,  0.005, -0.007, -0.020],
     [0.24, 0.078, 0.076, 0.061, 0.050, 0.040, 0.028, 0.009, -0.005, -0.020, -0.036],
     [0.30, 0.078, 0.076, 0.061, 0.050, 0.040, 0.028, 0.004, -0.015, -0.037, -0.060]])


BUFTTable = InterpND(method='lagrange2', points=(BUFT[1:, 0], BUFT[0, 1:]), values=BUFT[1:, 1:],
                     extrapolate=True)



import unittest
import os

import numpy as np
import openmdao.api as om
from openmdao.utils.assert_utils import (assert_check_partials,
                                         assert_near_equal)

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.mission.gasp_based.phases.breguet import RangeComp
from aviary.variable_info.variables import Dynamic


class TestBreguetResults(unittest.TestCase):
    def setUp(self):
        nn = 10

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "range_comp", RangeComp(num_nodes=nn), promotes=["*"])

        self.prob.setup(check=False, force_alloc_complex=True)

        self.prob.set_val("TAS_cruise", 458.8, units="kn")
        self.prob.set_val("mass", np.linspace(171481, 171481 - 10000, nn), units="lbm")
        self.prob.set_val(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, -
                          5870 * np.ones(nn,), units="lbm/h")

    def test_case1(self):

        tol = 1e-6
        self.prob.run_model()

        cruise_range = self.prob.get_val("cruise_range", units="NM")
        cruise_time = self.prob.get_val("cruise_time", units="s")

        t_expected = 6134.7240144
        r_expected = 781.83848643

        assert_near_equal(cruise_range[-1, ...], r_expected, tolerance=0.001)
        assert_near_equal(cruise_time[-1, ...], t_expected, tolerance=0.001)

        with np.printoptions(linewidth=1024):
            self.prob.model.list_outputs(prom_name=True, print_arrays=True)
            partial_data = self.prob.check_partials(method="cs")  # , out_stream=None)
        assert_check_partials(partial_data, atol=tol, rtol=tol)


class TestBreguetPartials(unittest.TestCase):
    def setUp(self):
        nn = 10

        self.prob = om.Problem()
        self.prob.model.add_subsystem(
            "range_comp", RangeComp(num_nodes=nn), promotes=["*"])

        self.prob.model.set_input_defaults(
            "TAS_cruise", 458.8 + 50 * np.random.rand(nn,), units="kn")
        self.prob.model.set_input_defaults(
            "mass", np.linspace(171481, 171481 - 10000, nn), units="lbm")
        self.prob.model.set_input_defaults(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, -5870 * np.ones(nn,), units="lbm/h")

        self.prob.setup(check=False, force_alloc_complex=True)

    def test_partials(self):

        tol = 1e-12
        self.prob.run_model()

        # cruise_range = self.prob.get_val("cruise_range", units="NM")
        # cruise_time = self.prob.get_val("cruise_time", units="s")

        # t_expected = 6134.7240144
        # r_expected = 781.83848643
        #
        # assert_near_equal(cruise_range[-1, ...], r_expected, tolerance=0.001)
        # assert_near_equal(cruise_time[-1, ...], t_expected, tolerance=0.001)

        with np.printoptions(linewidth=1024):
            self.prob.model.list_outputs(prom_name=True, print_arrays=True)
            partial_data = self.prob.check_partials(method="cs")  # , out_stream=None)
        assert_check_partials(partial_data, atol=tol, rtol=tol)


class TestBreguetResults(unittest.TestCase):
    def setUp(self):
        self.nn = nn = 100

        self.prob = om.Problem()
        self.prob.model.add_subsystem("range_comp",
                                      RangeComp(num_nodes=nn), promotes=["*"])

        self.prob.setup(check=False, force_alloc_complex=True)

        self.prob.set_val("TAS_cruise", 458.8, units="kn")
        self.prob.set_val("mass", np.linspace(171481, 171481 - 10000, nn), units="lbm")
        self.prob.set_val(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, -
                          5870 * np.ones(nn,), units="lbm/h")

    def test_results(self):
        self.prob.run_model()

        W = self.prob.get_val("mass", units="lbm") * GRAV_ENGLISH_LBM
        V = self.prob.get_val("TAS_cruise", units="kn")
        r = self.prob.get_val("cruise_range", units="NM")
        t = self.prob.get_val("cruise_time", units="h")
        fuel_flow = - \
            self.prob.get_val(
                Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, units="lbm/h")

        v_avg = (V[:-1] + V[1:])/2
        fuel_flow_avg = (fuel_flow[:-1] + fuel_flow[1:])/2

        # Range should be equal to the product of initial speed in the segment and change in time
        assert_near_equal(np.diff(r), v_avg * np.diff(t), tolerance=1.0E-5)

        # time should satisfy: dt = -dW / fuel_flow
        assert_near_equal(np.diff(t), -np.diff(W) / fuel_flow_avg, tolerance=1.0E-6)


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import openmdao
import openmdao.api as om
from openmdao.utils.assert_utils import assert_check_partials, assert_near_equal
from packaging import version

from aviary.subsystems.propulsion.engine_deck import EngineDeck
from aviary.subsystems.propulsion.propulsion_mission import (
    PropulsionMission, PropulsionSum)
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.preprocessors import preprocess_propulsion
from aviary.utils.functions import get_path
from aviary.validation_cases.validation_tests import get_flops_inputs
from aviary.variable_info.variables import Aircraft, Dynamic, Mission


class PropulsionMissionTest(unittest.TestCase):
    def setUp(self):
        self.prob = om.Problem()

    @unittest.skipIf(version.parse(openmdao.__version__) < version.parse("3.26"), "Skipping due to OpenMDAO version being too low (<3.26)")
    def test_case_1(self):
        # 'clean' test using GASP-derived engine deck
        nn = 20

        filename = get_path(
            'models/engines/turbofan_24k_1.deck')

        options = AviaryValues()
        options.set_val(Aircraft.Engine.DATA_FILE, filename)
        options.set_val(Aircraft.Engine.NUM_ENGINES, 2)
        options.set_val(Aircraft.Engine.SUBSONIC_FUEL_FLOW_SCALER, 1.0)
        options.set_val(Aircraft.Engine.SUPERSONIC_FUEL_FLOW_SCALER, 1.0)
        options.set_val(Aircraft.Engine.FUEL_FLOW_SCALER_CONSTANT_TERM, 0.0)
        options.set_val(Aircraft.Engine.FUEL_FLOW_SCALER_LINEAR_TERM, 1.0)
        options.set_val(Aircraft.Engine.CONSTANT_FUEL_CONSUMPTION, 0.0, units='lbm/h')
        options.set_val(Aircraft.Engine.SCALE_PERFORMANCE, True)
        options.set_val(Mission.Summary.FUEL_FLOW_SCALER, 1.0)
        options.set_val(Aircraft.Engine.SCALE_FACTOR, 0.5)
        options.set_val(Aircraft.Engine.GENERATE_FLIGHT_IDLE, False)
        options.set_val(Aircraft.Engine.IGNORE_NEGATIVE_THRUST, False)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_THRUST_FRACTION, 0.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MAX_FRACTION, 1.0)
        options.set_val(Aircraft.Engine.FLIGHT_IDLE_MIN_FRACTION, 0.08)
        options.set_val(Aircraft.Engine.GEOPOTENTIAL_ALT, False)
        options.set_val(Aircraft.Engine.INTERPOLATION_METHOD, 'slinear')

        engine = EngineDeck(options=options)
        preprocess_propulsion(options, [engine])

        self.prob.model = PropulsionMission(num_nodes=nn, aviary_options=options)

        IVC = om.IndepVarComp(Dynamic.Mission.MACH,
                              np.linspace(0, 0.8, nn),
                              units='unitless')
        IVC.add_output(Dynamic.Mission.ALTITUDE,
                       np.linspace(0, 40000, nn),
                       units='ft')
        IVC.add_output(Dynamic.Mission.THROTTLE,
                       np.linspace(1, 0.7, nn),
                       units='unitless')
        self.prob.model.add_subsystem('IVC', IVC, promotes=['*'])

        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val(Aircraft.Engine.SCALE_FACTOR, options.get_val(
            Aircraft.Engine.SCALE_FACTOR), units='unitless')

        self.prob.run_model()

        thrust = self.prob.get_val(Dynamic.Mission.THRUST_TOTAL, units='lbf')
        fuel_flow = self.prob.get_val(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, units='lbm/h')

        expected_thrust = np.array([26559.90955398, 24186.4637312, 21938.65874407,
                                    19715.77939805, 17507.00655484, 15461.29892872,
                                    13781.56317005, 12281.64477782, 10975.64977233,
                                    9457.34056514,  7994.85977229,  7398.22905691,
                                    7147.50679938,  6430.71565916,  5774.57932944,
                                    5165.15558103,  4583.1380952,  3991.15088149,
                                    3338.98524687, 2733.56788119])

        expected_fuel_flow = np.array([-14707.1792863, -14065.2831058, -13383.11681516,
                                       -12535.21693425, -11524.37848035, -10514.44342419,
                                       -9697.03653898,  -8936.66146966, -8203.85487648,
                                       -8447.54167564,  -8705.14277314,  -7470.29404109,
                                       -5980.15247732,  -5493.23821772,  -5071.79842346,
                                       -4660.12833977,  -4260.89619679,  -3822.61002621,
                                       -3344.41332545,  -2889.68646353])

        assert_near_equal(thrust, expected_thrust, tolerance=1e-10)
        assert_near_equal(fuel_flow, expected_fuel_flow, tolerance=1e-10)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)

    def test_propulsion_sum(self):
        nn = 2
        options = AviaryValues()
        options.set_val(Aircraft.Engine.NUM_ENGINES, np.array([3, 2]))
        # it doesn't matter what goes in engine models, as long as it is length 2
        options.set_val('engine_models', [1, 1])
        self.prob.model = om.Group()
        self.prob.model.add_subsystem('propsum',
                                      PropulsionSum(num_nodes=nn,
                                                    aviary_options=options),
                                      promotes=['*'])

        self.prob.setup(force_alloc_complex=True)

        self.prob.set_val(Dynamic.Mission.THRUST, np.array(
            [[500.4, 423.001], [325, 6780]]))
        self.prob.set_val(Dynamic.Mission.THRUST_MAX,
                          np.array([[602.11, 3554], [100, 9000]]))
        self.prob.set_val(Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE,
                          np.array([[123, -221.44], [-765.2, -1]]))
        self.prob.set_val(Dynamic.Mission.ELECTRIC_POWER,
                          np.array([[3.01, -12], [484.2, 8123]]))
        self.prob.set_val(Dynamic.Mission.NOX_RATE,
                          np.array([[322, 4610], [1.54, 2.844]]))

        self.prob.run_model()

        thrust = self.prob.get_val(Dynamic.Mission.THRUST_TOTAL, units='lbf')
        thrust_max = self.prob.get_val(Dynamic.Mission.THRUST_MAX_TOTAL, units='lbf')
        fuel_flow = self.prob.get_val(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, units='lb/h')
        electric_power = self.prob.get_val(
            Dynamic.Mission.ELECTRIC_POWER_TOTAL, units='kW')
        nox = self.prob.get_val(Dynamic.Mission.NOX_RATE_TOTAL, units='lb/h')

        expected_thrust = np.array([2347.202, 14535])
        expected_thrust_max = np.array([8914.33, 18300])
        expected_fuel_flow = np.array([-73.88, -2297.6])
        expected_electric_power = np.array([-14.97, 17698.6])
        expected_nox = np.array([10186, 10.308])

        assert_near_equal(thrust, expected_thrust, tolerance=1e-12)
        assert_near_equal(thrust_max, expected_thrust_max, tolerance=1e-12)
        assert_near_equal(fuel_flow, expected_fuel_flow, tolerance=1e-12)
        assert_near_equal(electric_power, expected_electric_power, tolerance=1e-12)
        assert_near_equal(nox, expected_nox, tolerance=1e-12)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)

    def test_case_2(self):
        # takes the large single aisle 2 test case and add a second set of engines to test summation
        nn = 20

        options = get_flops_inputs('LargeSingleAisle2FLOPS')

        engine = options.get_val('engine_models')[0]
        engine2 = options.deepcopy().get_val('engine_models')[0]
        engine2.name = 'engine2'
        preprocess_propulsion(options, [engine, engine2])

        self.prob.model = PropulsionMission(num_nodes=20, aviary_options=options)

        self.prob.model.add_subsystem(Dynamic.Mission.MACH,
                                      om.IndepVarComp(Dynamic.Mission.MACH,
                                                      np.linspace(0, 0.85, nn),
                                                      units='unitless'),
                                      promotes=['*'])

        self.prob.model.add_subsystem(
            Dynamic.Mission.ALTITUDE,
            om.IndepVarComp(
                Dynamic.Mission.ALTITUDE,
                np.linspace(0, 40000, nn),
                units='ft'),
            promotes=['*'])
        throttle = np.linspace(1.0, 0.6, nn)
        self.prob.model.add_subsystem(
            Dynamic.Mission.THROTTLE, om.IndepVarComp(Dynamic.Mission.THROTTLE, np.vstack((throttle, throttle)).transpose(), units='unitless'), promotes=['*'])

        self.prob.setup(force_alloc_complex=True)
        self.prob.set_val(Aircraft.Engine.SCALE_FACTOR, [0.975], units='unitless')

        self.prob.run_model()

        thrust = self.prob.get_val(Dynamic.Mission.THRUST_TOTAL, units='lbf')
        fuel_flow = self.prob.get_val(
            Dynamic.Mission.FUEL_FLOW_RATE_NEGATIVE_TOTAL, units='lbm/h')
        nox_rate = self.prob.get_val(Dynamic.Mission.NOX_RATE_TOTAL, units='lbm/h')

        expected_thrust = np.array([103583.64726051,  92899.15059987,  82826.62014006,  73006.74478288,
                                    63491.73778033,  55213.71927899,  48317.05801159,  42277.98362824,
                                    36870.43915515,  29716.58670587,  26271.29434561,  24680.25359966,
                                    22043.65303425,  19221.1253513,  16754.1861966,   14405.43665682,
                                    12272.31373152,  10141.72397926,   7869.3816548,    5792.62871788])

        expected_fuel_flow = np.array([-38238.66614438, -36078.76817864, -33777.65206416, -31057.41872898,
                                       -28036.92997813, -25279.48301301, -22902.98616678, -20749.08916211,
                                       -19058.23299911, -19972.32193796, -17701.86829646, -14370.68121827,
                                       -12584.1724091,  -11320.06786905, -10192.11938107,  -9100.08365082,
                                       -8100.4835652,   -7069.62950088,  -5965.78834865,  -4914.94081538])

        expected_nox_rate = np.array(
            [0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])

        assert_near_equal(thrust, expected_thrust, tolerance=1e-10)
        assert_near_equal(fuel_flow, expected_fuel_flow, tolerance=1e-10)
        assert_near_equal(nox_rate, expected_nox_rate, tolerance=1e-9)

        partial_data = self.prob.check_partials(out_stream=None, method="cs")
        assert_check_partials(partial_data, atol=1e-10, rtol=1e-10)


if __name__ == "__main__":
    unittest.main()



import inspect


def add_subsystem_variables_to_phase(phase, phase_name, external_subsystems):
    """
    Add subsystem variables like states, controls, parameters, and constraints to a phase.

    Parameters
    ----------
    phase : object
        The phase object to which variables are to be added.
    phase_name : str
        The name of the phase for which the variables are added. Useful for cases
        where the subsystem behavior varies based on phase name.
    external_subsystems : list
        List of external subsystem objects that contain variables to add to the phase.

    Returns
    -------
    phase : object
        The modified phase object with added variables.

    """

    # Loop through each subsystem in the list of external_subsystems
    for subsystem in external_subsystems:

        # Fetch the states from the current subsystem
        subsystem_states = subsystem.get_states()

        # Add each state and its corresponding arguments to the phase
        for state_name in subsystem_states:
            kwargs = subsystem_states[state_name]
            phase.add_state(state_name, **kwargs)

        # Check if 'get_controls' function in the subsystem accepts 'phase_name' as an argument
        arg_spec = inspect.getfullargspec(subsystem.get_controls)
        if 'phase_name' in arg_spec.args:
            controls = subsystem.get_controls(phase_name=phase_name)
        else:
            controls = subsystem.get_controls()

        for control_name in controls:
            kwargs = controls[control_name]
            phase.add_control(control_name, **kwargs)

        constraints = subsystem.get_constraints()

        # Add each constraint and its corresponding arguments to the phase
        for constraint_name in constraints:
            kwargs = constraints[constraint_name]
            if kwargs['type'] == 'boundary':
                kwargs.pop('type')
                phase.add_boundary_constraint(constraint_name, **kwargs)
            elif kwargs['type'] == 'path':
                kwargs.pop('type')
                phase.add_path_constraint(constraint_name, **kwargs)

    return phase


def get_initial(status, key, status_for_this_variable=False):
    # Check if status is a dictionary.
    # If so, return the value corresponding to the key or status_for_this_variable if the key is not found.
    # If not, return the value of status.
    if isinstance(status, dict):
        if key in status:
            status_for_this_variable = status[key]
    elif isinstance(status, bool):
        status_for_this_variable = status
    return status_for_this_variable



import openmdao.api as om

from aviary.utils.aviary_values import AviaryValues
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft


class StrutCalcs(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.SPAN, val=0)

        if self.options["aviary_options"].get_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            add_aviary_input(self, Aircraft.Strut.ATTACHMENT_LOCATION, val=0)
            add_aviary_output(
                self, Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=0)
        else:
            add_aviary_input(
                self, Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, val=0)
            add_aviary_output(self, Aircraft.Strut.ATTACHMENT_LOCATION, val=0)

    def setup_partials(self):

        if self.options["aviary_options"].get_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            self.declare_partials(
                Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, [Aircraft.Strut.ATTACHMENT_LOCATION, Aircraft.Wing.SPAN])
        else:
            self.declare_partials(
                Aircraft.Strut.ATTACHMENT_LOCATION, [Aircraft.Strut.ATTACHMENT_LOCATION_DIMENSIONLESS, Aircraft.Wing.SPAN])

    def compute(self, inputs, outputs):
        wing_span = inputs[Aircraft.Wing.SPAN]
        strut_loc_name = Aircraft.Strut.ATTACHMENT_LOCATION

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            if self.options["aviary_options"].get_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
                strut_x = inputs[strut_loc_name]
                outputs[strut_loc_name+"_dimensionless"] = strut_x / wing_span
            else:
                strut_x = inputs[strut_loc_name+"_dimensionless"]
                outputs[strut_loc_name] = strut_x * wing_span

    def compute_partials(self, inputs, partials):
        wing_span = inputs[Aircraft.Wing.SPAN]
        strut_loc_name = Aircraft.Strut.ATTACHMENT_LOCATION

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            if self.options["aviary_options"].get_val(Aircraft.Strut.DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
                partials[strut_loc_name+"_dimensionless", strut_loc_name] = 1 / wing_span
                partials[strut_loc_name+"_dimensionless",
                         Aircraft.Wing.SPAN] = - inputs[strut_loc_name] / wing_span**2
            else:
                partials[strut_loc_name, strut_loc_name+"_dimensionless"] = wing_span
                partials[strut_loc_name,
                         Aircraft.Wing.SPAN] = inputs[strut_loc_name+"_dimensionless"]


class FoldCalcs(om.ExplicitComponent):
    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):
        add_aviary_input(self, Aircraft.Wing.SPAN, val=0)

        if self.options["aviary_options"].get_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            add_aviary_input(self, Aircraft.Wing.FOLDED_SPAN, val=0)
            add_aviary_output(self, Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, val=0)
        else:
            add_aviary_input(self, Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, val=0)
            add_aviary_output(self, Aircraft.Wing.FOLDED_SPAN, val=0)

    def setup_partials(self):

        if self.options["aviary_options"].get_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            self.declare_partials(
                Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, [Aircraft.Wing.FOLDED_SPAN, Aircraft.Wing.SPAN])
        else:
            self.declare_partials(
                Aircraft.Wing.FOLDED_SPAN, [Aircraft.Wing.FOLDED_SPAN_DIMENSIONLESS, Aircraft.Wing.SPAN])

    def compute(self, inputs, outputs):
        wing_span = inputs[Aircraft.Wing.SPAN]
        folded_span_name = Aircraft.Wing.FOLDED_SPAN

        if self.options["aviary_options"].get_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            fold_y = inputs[folded_span_name]
            outputs[folded_span_name+"_dimensionless"] = fold_y / wing_span
        else:
            fold_y = inputs[folded_span_name+"_dimensionless"]
            outputs[folded_span_name] = fold_y * wing_span

    def compute_partials(self, inputs, partials):
        wing_span = inputs[Aircraft.Wing.SPAN]
        folded_span_name = Aircraft.Wing.FOLDED_SPAN

        if self.options["aviary_options"].get_val(Aircraft.Wing.FOLD_DIMENSIONAL_LOCATION_SPECIFIED, units='unitless'):
            partials[folded_span_name+"_dimensionless", folded_span_name] = 1 / wing_span
            partials[folded_span_name+"_dimensionless", Aircraft.Wing.SPAN] = - \
                inputs[folded_span_name] / (wing_span**2)
        else:
            partials[folded_span_name, folded_span_name+"_dimensionless"] = wing_span
            partials[folded_span_name,
                     Aircraft.Wing.SPAN] = inputs[folded_span_name+"_dimensionless"]


class DimensionalNonDimensionalInterchange(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        if aviary_options.get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            self.add_subsystem(
                "strut_calcs",
                StrutCalcs(aviary_options=aviary_options,),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"],
            )

        if aviary_options.get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            self.add_subsystem(
                "fold_calcs",
                FoldCalcs(aviary_options=aviary_options,),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"],
            )



import numpy as np
import openmdao.api as om

from aviary.constants import GRAV_ENGLISH_LBM
from aviary.subsystems.geometry.gasp_based.non_dimensional_conversion import \
    DimensionalNonDimensionalInterchange
from aviary.subsystems.geometry.gasp_based.strut import StrutGeom
from aviary.utils.aviary_values import AviaryValues
from aviary.utils.conflict_checks import check_fold_location_definition
from aviary.variable_info.functions import add_aviary_input, add_aviary_output
from aviary.variable_info.variables import Aircraft, Mission


class WingSize(om.ExplicitComponent):

    def initialize(self):
        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Mission.Design.GROSS_MASS, val=152000)
        add_aviary_input(self, Aircraft.Wing.LOADING, val=128)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)

        add_aviary_output(self, Aircraft.Wing.AREA, val=0)
        add_aviary_output(self, Aircraft.Wing.SPAN, val=0)

        self.declare_partials(
            Aircraft.Wing.AREA, [Mission.Design.GROSS_MASS, Aircraft.Wing.LOADING]
        )
        self.declare_partials(
            Aircraft.Wing.SPAN,
            [
                Aircraft.Wing.ASPECT_RATIO,
                Mission.Design.GROSS_MASS,
                Aircraft.Wing.LOADING,
            ],
        )

    def compute(self, inputs, outputs):

        gross_mass_initial = inputs[Mission.Design.GROSS_MASS]
        wing_loading = inputs[Aircraft.Wing.LOADING]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]

        wing_area = gross_mass_initial * GRAV_ENGLISH_LBM / wing_loading
        wingspan = (AR * wing_area) ** 0.5

        outputs[Aircraft.Wing.AREA] = wing_area
        outputs[Aircraft.Wing.SPAN] = wingspan

    def compute_partials(self, inputs, J):

        gross_mass_initial = inputs[Mission.Design.GROSS_MASS]
        wing_loading = inputs[Aircraft.Wing.LOADING]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]

        wing_area = gross_mass_initial * GRAV_ENGLISH_LBM / wing_loading

        J[Aircraft.Wing.AREA,
            Mission.Design.GROSS_MASS] = dWA_dGMT = GRAV_ENGLISH_LBM / wing_loading
        J[Aircraft.Wing.AREA, Aircraft.Wing.LOADING] = dWA_dWL = (
            -gross_mass_initial * GRAV_ENGLISH_LBM / wing_loading**2
        )

        J[Aircraft.Wing.SPAN, Aircraft.Wing.ASPECT_RATIO] = (
            0.5 * wing_area**0.5 * AR ** (-0.5)
        )
        J[Aircraft.Wing.SPAN, Mission.Design.GROSS_MASS] = (
            0.5 * AR**0.5 * wing_area ** (-0.5) * dWA_dGMT
        )
        J[Aircraft.Wing.SPAN, Aircraft.Wing.LOADING] = (
            0.5 * AR**0.5 * wing_area ** (-0.5) * dWA_dWL
        )


class WingParameters(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        add_aviary_input(self, Aircraft.Wing.AREA, val=2)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=2)
        add_aviary_input(self, Aircraft.Wing.ASPECT_RATIO, val=10.13)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)
        add_aviary_input(self, Aircraft.Wing.SWEEP, val=25)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.11)
        add_aviary_input(self, Aircraft.Fuselage.AVG_DIAMETER, val=10)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_TIP, val=0.1)

        if not self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):

            add_aviary_input(self, Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6)
            add_aviary_output(self, Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=0)

            self.declare_partials(
                Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                [
                    Aircraft.Fuel.WING_FUEL_FRACTION,
                    Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                    Aircraft.Fuselage.AVG_DIAMETER,
                    Aircraft.Wing.SPAN,
                    Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
                    Aircraft.Wing.AREA,
                    Aircraft.Wing.TAPER_RATIO,
                    Aircraft.Wing.ASPECT_RATIO,
                ],
            )

        add_aviary_output(self, Aircraft.Wing.CENTER_CHORD, val=0)
        add_aviary_output(self, Aircraft.Wing.AVERAGE_CHORD, val=0)
        add_aviary_output(self, Aircraft.Wing.ROOT_CHORD, val=0)
        add_aviary_output(self, Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, val=0)
        add_aviary_output(self, Aircraft.Wing.LEADING_EDGE_SWEEP,
                          val=0.4763948, units="rad")

        self.declare_partials(
            Aircraft.Wing.CENTER_CHORD,
            [Aircraft.Wing.AREA, Aircraft.Wing.SPAN, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            Aircraft.Wing.AVERAGE_CHORD,
            [Aircraft.Wing.AREA, Aircraft.Wing.SPAN, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            [
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
                Aircraft.Wing.TAPER_RATIO,
            ],
        )
        self.declare_partials(
            Aircraft.Wing.ROOT_CHORD,
            [
                Aircraft.Wing.AREA,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.TAPER_RATIO,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Fuselage.AVG_DIAMETER,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
            ],
        )
        self.declare_partials(
            Aircraft.Wing.LEADING_EDGE_SWEEP,
            [
                Aircraft.Wing.TAPER_RATIO,
                Aircraft.Wing.ASPECT_RATIO,
                Aircraft.Wing.SWEEP,
            ],
        )

    def compute(self, inputs, outputs):

        wing_area = inputs[Aircraft.Wing.AREA]
        wingspan = inputs[Aircraft.Wing.SPAN]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        sweep_c4 = inputs[Aircraft.Wing.SWEEP]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        tc_ratio_tip = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_TIP]

        center_chord = 2.0 * wing_area / wingspan / (1.0 + taper_ratio)
        avg_chord = (2.0 * center_chord / 3.0) * (
            (1.0 + taper_ratio) - (taper_ratio / (1.0 + taper_ratio))
        )
        tan_sweep_LE = (1.0 - taper_ratio) / (1.0 + taper_ratio) / AR + np.tan(
            sweep_c4 * np.pi / 180.0
        )
        outputs[Aircraft.Wing.LEADING_EDGE_SWEEP] = np.arctan(tan_sweep_LE)
        tan_sweep_TE = 3.0 * (taper_ratio - 1.0) / (1.0 + taper_ratio) / AR + np.tan(
            sweep_c4 * (np.pi / 180)
        )
        sweep_TE = np.arctan(tan_sweep_TE)
        FHP = (
            2.0
            * (
                tc_ratio_root
                * center_chord
                * (cabin_width - (tc_ratio_root * center_chord))
            )
            ** 0.5
            + 0.4
        )
        HP = FHP * tan_sweep_LE / 2.0
        root_chord = center_chord - HP + FHP * tan_sweep_TE / 2.0
        tc_ratio_avg = (
            (tc_ratio_root - cabin_width / wingspan * (tc_ratio_root - tc_ratio_tip))
            * (1.0 - cabin_width / wingspan * (1.0 - taper_ratio))
            + taper_ratio * tc_ratio_tip
        ) / (1.0 + taper_ratio - cabin_width / wingspan * (1.0 - taper_ratio))

        outputs[Aircraft.Wing.CENTER_CHORD] = center_chord
        outputs[Aircraft.Wing.AVERAGE_CHORD] = avg_chord
        outputs[Aircraft.Wing.ROOT_CHORD] = root_chord
        outputs[Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED] = tc_ratio_avg

        if not self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]

            geometric_fuel_vol = (
                fuel_vol_frac
                * 0.888889
                * tc_ratio_avg
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            outputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = geometric_fuel_vol

    def compute_partials(self, inputs, J):

        wing_area = inputs[Aircraft.Wing.AREA]
        wingspan = inputs[Aircraft.Wing.SPAN]
        AR = inputs[Aircraft.Wing.ASPECT_RATIO]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        sweep_c4 = inputs[Aircraft.Wing.SWEEP]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        cabin_width = inputs[Aircraft.Fuselage.AVG_DIAMETER]
        tc_ratio_tip = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_TIP]

        center_chord = 2.0 * wing_area / wingspan / (1.0 + taper_ratio)
        FHP = (
            2.0
            * (
                tc_ratio_root
                * center_chord
                * (cabin_width - (tc_ratio_root * center_chord))
            )
            ** 0.5
            + 0.4
        )

        tan_sweep_LE = (1.0 - taper_ratio) / (1.0 + taper_ratio) / AR + np.tan(
            sweep_c4 * (np.pi / 180)
        )
        tan_sweep_TE = 3.0 * (taper_ratio - 1.0) / (1.0 + taper_ratio) / AR + np.tan(
            sweep_c4 * (np.pi / 180)
        )

        dCenterChord_dWingArea = 2 / (wingspan * (1 + taper_ratio))
        dCenterChord_dWingspan = -2 * wing_area / ((1 + taper_ratio) * wingspan**2)
        dCenterChord_dTaperRatio = -2 * wing_area / (wingspan * (1 + taper_ratio) ** 2)
        dFHP_dTcRatioRoot = (
            tc_ratio_root * center_chord * (cabin_width - tc_ratio_root * center_chord)
        ) ** (-0.5) * (
            center_chord * cabin_width - 2 * tc_ratio_root * center_chord**2
        )
        dFHP_dCabinWidth = (
            (
                tc_ratio_root
                * center_chord
                * (cabin_width - tc_ratio_root * center_chord)
            )
            ** (-0.5)
            * tc_ratio_root
            * center_chord
        )
        dFHP_dWingArea = (
            (
                tc_ratio_root
                * center_chord
                * (cabin_width - tc_ratio_root * center_chord)
            )
            ** (-0.5)
            * (tc_ratio_root * cabin_width - 2 * tc_ratio_root**2 * center_chord)
            * dCenterChord_dWingArea
        )
        dFHP_dWingspan = (
            (
                tc_ratio_root
                * center_chord
                * (cabin_width - tc_ratio_root * center_chord)
            )
            ** (-0.5)
            * (tc_ratio_root * cabin_width - 2 * tc_ratio_root**2 * center_chord)
            * dCenterChord_dWingspan
        )
        dFHP_dTaperRatio = (
            (
                tc_ratio_root
                * center_chord
                * (cabin_width - tc_ratio_root * center_chord)
            )
            ** (-0.5)
            * (tc_ratio_root * cabin_width - 2 * tc_ratio_root**2 * center_chord)
            * dCenterChord_dTaperRatio
        )

        dTanSweepLE_dTaperRatio = (-(1 + taper_ratio) * AR - (1 - taper_ratio) * AR) / (
            (1 + taper_ratio) ** 2 * AR**2
        )
        dTanSweepLE_dAR = -(1 - taper_ratio) / ((1 + taper_ratio) * AR**2)
        dTanSweepLE_dSweepC4 = dTanSweepTE_dSweepC4 = (
            (np.pi / 180) * 1 / np.cos(sweep_c4 * (np.pi / 180)) ** 2
        )
        dTanSweepTE_dTaperRatio = (
            3
            * ((1 + taper_ratio) * AR - (taper_ratio - 1) * AR)
            / ((1 + taper_ratio) ** 2 * AR**2)
        )
        dTanSweepTE_dAR = -3 * (taper_ratio - 1) / ((1 + taper_ratio) * AR**2)

        dRootChord_dWingArea = dCenterChord_dWingArea + dFHP_dWingArea * (
            -tan_sweep_LE / 2 + tan_sweep_TE / 2
        )
        dRootChord_dWingspan = dCenterChord_dWingspan + dFHP_dWingspan * (
            -tan_sweep_LE / 2 + tan_sweep_TE / 2
        )
        dRootChord_dTaperRatio = (
            dCenterChord_dTaperRatio
            + dFHP_dTaperRatio * (-tan_sweep_LE / 2 + tan_sweep_TE / 2)
            + FHP * (-dTanSweepLE_dTaperRatio / 2 + dTanSweepTE_dTaperRatio / 2)
        )
        dRootChord_dTcRatioRoot = dFHP_dTcRatioRoot * (
            -tan_sweep_LE / 2 + tan_sweep_TE / 2
        )
        dRootChord_dCabinWidth = dFHP_dCabinWidth * (
            -tan_sweep_LE / 2 + tan_sweep_TE / 2
        )
        dRootChord_dAR = FHP * (-dTanSweepLE_dAR / 2 + dTanSweepTE_dAR / 2)
        dRootChord_dSweepC4 = FHP * 0.5 * (-dTanSweepLE_dSweepC4 + dTanSweepTE_dSweepC4)

        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.AREA] = dRootChord_dWingArea
        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.SPAN] = dRootChord_dWingspan
        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.TAPER_RATIO] = dRootChord_dTaperRatio
        J[
            Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT
        ] = dRootChord_dTcRatioRoot
        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Fuselage.AVG_DIAMETER] = dRootChord_dCabinWidth
        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.ASPECT_RATIO] = dRootChord_dAR
        J[Aircraft.Wing.ROOT_CHORD, Aircraft.Wing.SWEEP] = dRootChord_dSweepC4

        J[Aircraft.Wing.CENTER_CHORD, Aircraft.Wing.AREA] = dCenterChord_dWingArea
        J[Aircraft.Wing.CENTER_CHORD, Aircraft.Wing.SPAN] = dCenterChord_dWingspan
        J[
            Aircraft.Wing.CENTER_CHORD, Aircraft.Wing.TAPER_RATIO
        ] = dCenterChord_dTaperRatio

        J[Aircraft.Wing.AVERAGE_CHORD, Aircraft.Wing.AREA] = (
            (2.0 / 3.0)
            * ((1.0 + taper_ratio) - (taper_ratio / (1.0 + taper_ratio)))
            * dCenterChord_dWingArea
        )
        J[Aircraft.Wing.AVERAGE_CHORD, Aircraft.Wing.SPAN] = (
            (2.0 / 3.0)
            * ((1.0 + taper_ratio) - (taper_ratio / (1.0 + taper_ratio)))
            * dCenterChord_dWingspan
        )
        J[
            Aircraft.Wing.AVERAGE_CHORD, Aircraft.Wing.TAPER_RATIO
        ] = dCenterChord_dTaperRatio * (2.0 / 3.0) * (
            (1.0 + taper_ratio) - (taper_ratio / (1.0 + taper_ratio))
        ) + (
            2.0 * center_chord / 3.0
        ) * (
            1 - ((1 + taper_ratio) - taper_ratio) / (1.0 + taper_ratio) ** 2
        )

        tc_ratio_avg = (
            (tc_ratio_root - cabin_width / wingspan * (tc_ratio_root - tc_ratio_tip))
            * (1.0 - cabin_width / wingspan * (1.0 - taper_ratio))
            + taper_ratio * tc_ratio_tip
        ) / (1.0 + taper_ratio - cabin_width / wingspan * (1.0 - taper_ratio))
        a = tc_ratio_root - cabin_width / wingspan * (tc_ratio_root - tc_ratio_tip)
        b = 1.0 - cabin_width / wingspan * (1.0 - taper_ratio)
        c = taper_ratio * tc_ratio_tip
        d = 1.0 + taper_ratio - cabin_width / wingspan * (1.0 - taper_ratio)

        dAB_dCabW = (
            a * (taper_ratio - 1) / wingspan
            + b * (tc_ratio_tip - tc_ratio_root) / wingspan
        )
        dD_dCabW = (taper_ratio - 1) / wingspan
        dAB_dWingspan = (
            a * cabin_width * (1 - taper_ratio) / wingspan**2
            + b * cabin_width * (tc_ratio_root - tc_ratio_tip) / wingspan**2
        )
        dD_dWingspan = cabin_width * (1 - taper_ratio) / wingspan**2
        dABC_dTR = a * cabin_width / wingspan + tc_ratio_tip
        dD_dTR = 1 + cabin_width / wingspan

        J[
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
        ] = dTCA_dTCR = (
            (1 - cabin_width / wingspan) * b / d
        )
        J[
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, Aircraft.Fuselage.AVG_DIAMETER
        ] = dTCA_dCabW = (d * dAB_dCabW - (a * b + c) * dD_dCabW) / d**2
        J[
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, Aircraft.Wing.SPAN
        ] = dTCA_dWingspan = (d * dAB_dWingspan - (a * b + c) * dD_dWingspan) / d**2
        J[
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED,
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
        ] = dTCA_dTCT = (cabin_width / wingspan * b + taper_ratio) / d
        J[
            Aircraft.Wing.THICKNESS_TO_CHORD_UNWEIGHTED, Aircraft.Wing.TAPER_RATIO
        ] = dTCA_dTR = (d * dABC_dTR - (a * b + c) * dD_dTR) / d**2

        trp1 = taper_ratio + 1
        swprad = np.pi * sweep_c4 / 180.0
        tswprad = np.tan(swprad)
        denom = AR**2 * trp1**2 + (AR * trp1 * tswprad - taper_ratio + 1) ** 2
        J[Aircraft.Wing.LEADING_EDGE_SWEEP, Aircraft.Wing.TAPER_RATIO] = -2 * AR / denom
        J[Aircraft.Wing.LEADING_EDGE_SWEEP, Aircraft.Wing.ASPECT_RATIO] = (
            (taper_ratio - 1) * trp1 / denom
        )
        J[Aircraft.Wing.LEADING_EDGE_SWEEP, Aircraft.Wing.SWEEP] = (
            np.pi * AR**2 * trp1**2 / denom / 180 / np.cos(swprad) ** 2
        )

        if not self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]
            geometric_fuel_vol = (
                fuel_vol_frac
                * 0.888889
                * tc_ratio_avg
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            num = (
                fuel_vol_frac
                * 0.888889
                * tc_ratio_avg
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            )
            den = (AR**0.5) * ((taper_ratio + 1.0) ** 2.0)
            dNum_dTR = (
                fuel_vol_frac
                * 0.888889
                * dTCA_dTR
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
                + fuel_vol_frac * 0.888889 * tc_ratio_avg * (wing_area**1.5) * 2
            )
            dDen_dTR = 2 * (AR**0.5) * (taper_ratio + 1.0)

            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Fuel.WING_FUEL_FRACTION] = (
                0.888889 * tc_ratio_avg * (wing_area**1.5) * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.AREA] = (
                1.5
                * (
                    fuel_vol_frac
                    * 0.888889
                    * tc_ratio_avg
                    * (wing_area**0.5)
                    * (2.0 * taper_ratio + 1.0)
                )
                / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            )
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.TAPER_RATIO] = (
                den * dNum_dTR - num * dDen_dTR
            ) / den**2
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.ASPECT_RATIO] = (
                -0.5
                * (
                    fuel_vol_frac
                    * 0.888889
                    * tc_ratio_avg
                    * (wing_area**1.5)
                    * (2.0 * taper_ratio + 1.0)
                )
                / ((AR**1.5) * (taper_ratio + 1.0) ** 2.0)
            )
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT] = (
                fuel_vol_frac
                * 0.888889
                * dTCA_dTCR
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Fuselage.AVG_DIAMETER] = (
                fuel_vol_frac
                * 0.888889
                * dTCA_dCabW
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.SPAN] = (
                fuel_vol_frac
                * 0.888889
                * dTCA_dWingspan
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))
            J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.THICKNESS_TO_CHORD_TIP] = (
                fuel_vol_frac
                * 0.888889
                * dTCA_dTCT
                * (wing_area**1.5)
                * (2.0 * taper_ratio + 1.0)
            ) / ((AR**0.5) * ((taper_ratio + 1.0) ** 2.0))


class WingFold(om.ExplicitComponent):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        if not self.options["aviary_options"].get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, units='unitless'):
            self.add_input(
                "strut_y",
                val=25,
                units="ft",
                desc="YSTRUT: attachment location of strut",
            )

            self.declare_partials("nonfolded_taper_ratio", "strut_y")
            self.declare_partials(Aircraft.Wing.FOLDING_AREA, "strut_y")
            self.declare_partials("nonfolded_wing_area", "strut_y")
            self.declare_partials("tc_ratio_mean_folded", "strut_y")
            self.declare_partials("nonfolded_AR", "strut_y")
            self.declare_partials(Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, "strut_y")

        else:
            self.declare_partials("nonfolded_taper_ratio", Aircraft.Wing.FOLDED_SPAN)
            self.declare_partials(Aircraft.Wing.FOLDING_AREA, Aircraft.Wing.FOLDED_SPAN)
            self.declare_partials("nonfolded_wing_area", Aircraft.Wing.FOLDED_SPAN)
            self.declare_partials("tc_ratio_mean_folded", Aircraft.Wing.FOLDED_SPAN)
            self.declare_partials("nonfolded_AR", Aircraft.Wing.FOLDED_SPAN)
            self.declare_partials(Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
                                  Aircraft.Wing.FOLDED_SPAN)

            add_aviary_input(self, Aircraft.Wing.FOLDED_SPAN, val=25, units='ft')

        add_aviary_input(self, Aircraft.Wing.AREA, val=200)
        add_aviary_input(self, Aircraft.Wing.SPAN, val=118)
        add_aviary_input(self, Aircraft.Wing.TAPER_RATIO, val=0.33)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.11)
        add_aviary_input(self, Aircraft.Wing.THICKNESS_TO_CHORD_TIP, val=0.1)
        add_aviary_input(self, Aircraft.Fuel.WING_FUEL_FRACTION, val=0.6)

        self.add_output(
            "nonfolded_taper_ratio",
            val=0.1,
            units="unitless",
            desc="SLM_NF: taper ratio between wing root and fold location",
        )

        add_aviary_output(self, Aircraft.Wing.FOLDING_AREA, val=50)

        self.add_output(
            "nonfolded_wing_area",
            val=150,
            units="ft**2",
            desc="SW_NF: wing area of part of wings that does not fold",
        )
        self.add_output(
            "tc_ratio_mean_folded",
            val=0.12,
            units="unitless",
            desc="TCM: mean value of thickess to chord ratio between root and fold",
        )
        self.add_output(
            "nonfolded_AR",
            val=10,
            units="unitless",
            desc="AR_NF: aspect ratio of non-folding part of wing",
        )

        add_aviary_output(self, Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, val=0)

        self.declare_partials(
            "nonfolded_taper_ratio",
            [Aircraft.Wing.AREA, Aircraft.Wing.SPAN, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            Aircraft.Wing.FOLDING_AREA,
            [Aircraft.Wing.SPAN, Aircraft.Wing.AREA, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            "nonfolded_wing_area",
            [Aircraft.Wing.AREA, Aircraft.Wing.SPAN, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            "tc_ratio_mean_folded",
            [
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
                Aircraft.Wing.SPAN,
            ],
        )
        self.declare_partials(
            "nonfolded_AR",
            [Aircraft.Wing.AREA, Aircraft.Wing.SPAN, Aircraft.Wing.TAPER_RATIO],
        )
        self.declare_partials(
            Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX,
            [
                Aircraft.Fuel.WING_FUEL_FRACTION,
                Aircraft.Wing.THICKNESS_TO_CHORD_ROOT,
                Aircraft.Wing.THICKNESS_TO_CHORD_TIP,
                Aircraft.Wing.SPAN,
                Aircraft.Wing.AREA,
                Aircraft.Wing.TAPER_RATIO,
            ],
        )

    def compute(self, inputs, outputs):

        wing_area = inputs[Aircraft.Wing.AREA]
        wingspan = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        tc_ratio_tip = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_TIP]
        fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]

        if not self.options["aviary_options"].get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, units='unitless'):

            strut_y = inputs["strut_y"]
            location = strut_y

        else:
            fold_y = inputs[Aircraft.Wing.FOLDED_SPAN]
            location = fold_y / 2.0

        root_chord_wing = 2 * wing_area / (wingspan * (1 + taper_ratio))
        tip_chord = taper_ratio * root_chord_wing
        fold_chord = root_chord_wing + location * (tip_chord - root_chord_wing) / (
            wingspan / 2.0
        )
        nonfolded_taper_ratio = fold_chord / root_chord_wing
        folding_area = (wingspan / 2.0 - location) * (fold_chord + tip_chord)

        nonfolded_wing_area = wing_area - folding_area

        if (wingspan / 2.0) < location:
            raise ValueError(
                "Error: The wingspan provided is less than the wingspan of the wing fold."
            )

        tc_ratio_fold = tc_ratio_root + location * (tc_ratio_tip - tc_ratio_root) / (
            wingspan / 2.0
        )
        tc_ratio_mean_folded = 0.5 * (tc_ratio_root + tc_ratio_fold)
        nonfolded_AR = 4.0 * location**2 / nonfolded_wing_area
        geometric_fuel_vol = (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * (nonfolded_wing_area**1.5)
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) / ((nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0))

        outputs["nonfolded_taper_ratio"] = nonfolded_taper_ratio
        outputs[Aircraft.Wing.FOLDING_AREA] = folding_area
        outputs["nonfolded_wing_area"] = nonfolded_wing_area
        outputs["tc_ratio_mean_folded"] = tc_ratio_mean_folded
        outputs["nonfolded_AR"] = nonfolded_AR
        outputs[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX] = geometric_fuel_vol

    def compute_partials(self, inputs, J):

        wing_area = inputs[Aircraft.Wing.AREA]
        wingspan = inputs[Aircraft.Wing.SPAN]
        taper_ratio = inputs[Aircraft.Wing.TAPER_RATIO]
        tc_ratio_root = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_ROOT]
        tc_ratio_tip = inputs[Aircraft.Wing.THICKNESS_TO_CHORD_TIP]
        fuel_vol_frac = inputs[Aircraft.Fuel.WING_FUEL_FRACTION]

        if not self.options["aviary_options"].get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, units='unitless'):

            strut_y = inputs["strut_y"]
            location = strut_y
            dLoc_dWingspan = 0
            dLoc_dy = 1
            wrt = "strut_y"

        else:
            fold_y = inputs[Aircraft.Wing.FOLDED_SPAN]
            wrt = Aircraft.Wing.FOLDED_SPAN

            location = fold_y / 2.0
            dLoc_dWingspan = 0
            dLoc_dy = 1 / 2

        root_chord_wing = 2 * wing_area / (wingspan * (1 + taper_ratio))
        tip_chord = taper_ratio * root_chord_wing
        fold_chord = root_chord_wing + location * (tip_chord - root_chord_wing) / (
            wingspan / 2.0
        )
        nonfolded_taper_ratio = fold_chord / root_chord_wing
        folding_area = (wingspan / 2.0 - location) * (fold_chord + tip_chord)
        nonfolded_wing_area = wing_area - folding_area
        tc_ratio_fold = tc_ratio_root + location * (tc_ratio_tip - tc_ratio_root) / (
            wingspan / 2.0
        )
        tc_ratio_mean_folded = 0.5 * (tc_ratio_root + tc_ratio_fold)
        nonfolded_AR = 4.0 * location**2 / nonfolded_wing_area

        dRootChordWing_dWingArea = 2 / (wingspan * (1 + taper_ratio))
        dRootChordWing_dWingspan = -2 * wing_area / (wingspan**2 * (1 + taper_ratio))
        dRootChordWing_dTaperRatio = (
            -2 * wing_area / (wingspan * (1 + taper_ratio) ** 2)
        )

        dTipChord_dTaperRatio = (
            taper_ratio * dRootChordWing_dTaperRatio + root_chord_wing
        )
        dTipChord_dWingArea = taper_ratio * dRootChordWing_dWingArea
        dTipChord_dWingspan = taper_ratio * dRootChordWing_dWingspan

        dFoldChord_dWingArea = dRootChordWing_dWingArea + location * (
            dTipChord_dWingArea - dRootChordWing_dWingArea
        ) / (wingspan / 2)
        dFoldChord_dWingspan = (
            dRootChordWing_dWingspan
            + location
            * (
                (wingspan / 2) * (dTipChord_dWingspan - dRootChordWing_dWingspan)
                - (tip_chord - root_chord_wing) * (1 / 2)
            )
            / (wingspan / 2) ** 2
            + dLoc_dWingspan * (tip_chord - root_chord_wing) / (wingspan / 2.0)
        )
        dFoldChord_dTaperRatio = dRootChordWing_dTaperRatio + location * (
            dTipChord_dTaperRatio - dRootChordWing_dTaperRatio
        ) / (wingspan / 2)
        dFoldChord_dy = dLoc_dy * (tip_chord - root_chord_wing) / (wingspan / 2.0)

        dTcRatioFold_dTCR = 1 + location * (-1) / (wingspan / 2)
        dTcRatioFold_dTCT = location / (wingspan / 2)
        dTcRatioFold_dWingspan = (
            (tc_ratio_tip - tc_ratio_root)
            * ((wingspan / 2) * dLoc_dWingspan - location * 0.5)
            / (wingspan / 2) ** 2
        )
        dTcRatioFold_dy = dLoc_dy * (tc_ratio_tip - tc_ratio_root) / (wingspan / 2.0)

        J["nonfolded_taper_ratio", Aircraft.Wing.AREA] = dNFTR_dWingArea = (
            root_chord_wing * dFoldChord_dWingArea
            - fold_chord * dRootChordWing_dWingArea
        ) / root_chord_wing**2
        J["nonfolded_taper_ratio", Aircraft.Wing.SPAN] = dNFTR_dWingspan = (
            root_chord_wing * dFoldChord_dWingspan
            - fold_chord * dRootChordWing_dWingspan
        ) / root_chord_wing**2
        J["nonfolded_taper_ratio", Aircraft.Wing.TAPER_RATIO] = dNFTR_dTaperRatio = (
            root_chord_wing * dFoldChord_dTaperRatio
            - fold_chord * dRootChordWing_dTaperRatio
        ) / root_chord_wing**2
        J["nonfolded_taper_ratio", wrt] = dNFTR_dy = dFoldChord_dy / root_chord_wing

        J[Aircraft.Wing.FOLDING_AREA, Aircraft.Wing.AREA] = dFoldingArea_dWingArea = (
            wingspan / 2.0 - location
        ) * (dFoldChord_dWingArea + dTipChord_dWingArea)
        J[Aircraft.Wing.FOLDING_AREA, Aircraft.Wing.SPAN] = dFoldingArea_dWingspan = (
            wingspan / 2.0 - location
        ) * (dFoldChord_dWingspan + dTipChord_dWingspan) + (fold_chord + tip_chord) * (
            0.5 - dLoc_dWingspan
        )
        J[
            Aircraft.Wing.FOLDING_AREA, Aircraft.Wing.TAPER_RATIO
        ] = dFoldingArea_dTaperRatio = (wingspan / 2.0 - location) * (
            dFoldChord_dTaperRatio + dTipChord_dTaperRatio
        )
        J[Aircraft.Wing.FOLDING_AREA, wrt] = dFoldingArea_dy = -dLoc_dy * (
            fold_chord + tip_chord
        ) + (wingspan / 2 - location) * dLoc_dy * (tip_chord - root_chord_wing) / (
            wingspan / 2.0
        )

        J["nonfolded_wing_area", Aircraft.Wing.AREA] = dNFWA_dWingArea = (
            1 - dFoldingArea_dWingArea
        )
        J[
            "nonfolded_wing_area", Aircraft.Wing.SPAN
        ] = dNFWA_dWingspan = -dFoldingArea_dWingspan
        J[
            "nonfolded_wing_area", Aircraft.Wing.TAPER_RATIO
        ] = dNFWA_dTaperRatio = -dFoldingArea_dTaperRatio
        J["nonfolded_wing_area", wrt] = dNFWA_dy = -dFoldingArea_dy

        J[
            "tc_ratio_mean_folded", Aircraft.Wing.THICKNESS_TO_CHORD_ROOT
        ] = dTCRMeanFolded_dTCR = 0.5 * (1 + dTcRatioFold_dTCR)
        J[
            "tc_ratio_mean_folded", Aircraft.Wing.THICKNESS_TO_CHORD_TIP
        ] = dTCRMeanFolded_dTCT = (0.5 * dTcRatioFold_dTCT)
        J["tc_ratio_mean_folded", Aircraft.Wing.SPAN] = dTCRMeanFolded_dWingspan = (
            0.5 * dTcRatioFold_dWingspan
        )
        J["tc_ratio_mean_folded", wrt] = dTCRMeanFolded_dy = 0.5 * dTcRatioFold_dy

        J["nonfolded_AR", Aircraft.Wing.AREA] = dNFAR_dWingArea = (
            -4 * location**2 / nonfolded_wing_area**2 * dNFWA_dWingArea
        )
        J["nonfolded_AR", Aircraft.Wing.SPAN] = dNFAR_dWingspan = (
            4
            * (
                nonfolded_wing_area * 2 * location * dLoc_dWingspan
                - location**2 * dNFWA_dWingspan
            )
            / nonfolded_wing_area**2
        )
        J["nonfolded_AR", Aircraft.Wing.TAPER_RATIO] = dNFAR_dTaperRatio = (
            -4 * location**2 / nonfolded_wing_area**2 * dNFWA_dTaperRatio
        )
        J["nonfolded_AR", wrt] = dNFAR_dy = (
            4
            * (nonfolded_wing_area * 2 * location * dLoc_dy - location**2 * dNFWA_dy)
            / nonfolded_wing_area**2
        )

        geometric_fuel_vol = (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * (nonfolded_wing_area**1.5)
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) / ((nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0))
        a = (nonfolded_wing_area**1.5) * (2.0 * nonfolded_taper_ratio + 1.0)
        num = fuel_vol_frac * 0.888889 * tc_ratio_mean_folded * a
        den = (nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0)

        dA_dWingspan = (
            1.5
            * nonfolded_wing_area**0.5
            * dNFWA_dWingspan
            * (2 * nonfolded_taper_ratio + 1)
            + nonfolded_wing_area**1.5 * 2 * dNFTR_dWingspan
        )
        dA_dy = (
            1.5
            * nonfolded_wing_area**0.5
            * dNFWA_dy
            * (2 * nonfolded_taper_ratio + 1)
            + nonfolded_wing_area**1.5 * 2 * dNFTR_dy
        )

        dNum_dWingspan = (
            fuel_vol_frac * 0.888889 * dTCRMeanFolded_dWingspan * a
            + fuel_vol_frac * 0.888889 * tc_ratio_mean_folded * dA_dWingspan
        )
        dDen_dWingspan = (
            0.5
            * (nonfolded_AR ** (-0.5))
            * dNFAR_dWingspan
            * ((nonfolded_taper_ratio + 1.0) ** 2.0)
            + ((nonfolded_AR**0.5) * 2 * ((nonfolded_taper_ratio + 1.0)))
            * dNFTR_dWingspan
        )
        dNum_dy = (
            fuel_vol_frac * 0.888889 * dTCRMeanFolded_dy * a
            + fuel_vol_frac * 0.888889 * tc_ratio_mean_folded * dA_dy
        )
        dDen_dy = (
            0.5
            * (nonfolded_AR ** (-0.5))
            * dNFAR_dy
            * ((nonfolded_taper_ratio + 1.0) ** 2.0)
            + ((nonfolded_AR**0.5) * 2 * ((nonfolded_taper_ratio + 1.0))) * dNFTR_dy
        )
        dNum_dWingArea = (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * 1.5
            * (nonfolded_wing_area**0.5)
            * dNFWA_dWingArea
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) + (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * (nonfolded_wing_area**1.5)
            * 2
        ) * dNFTR_dWingArea
        dDen_dWingArea = (
            0.5
            * (nonfolded_AR ** (-0.5))
            * dNFAR_dWingArea
            * ((nonfolded_taper_ratio + 1.0) ** 2.0)
            + ((nonfolded_AR**0.5) * 2 * ((nonfolded_taper_ratio + 1.0)))
            * dNFTR_dWingArea
        )
        dNum_dTaperRatio = (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * 1.5
            * (nonfolded_wing_area**0.5)
            * dNFWA_dTaperRatio
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) + (
            fuel_vol_frac
            * 0.888889
            * tc_ratio_mean_folded
            * (nonfolded_wing_area**1.5)
            * 2
        ) * dNFTR_dTaperRatio
        dDen_dTaperRatio = (
            0.5
            * (nonfolded_AR ** (-0.5))
            * dNFAR_dTaperRatio
            * ((nonfolded_taper_ratio + 1.0) ** 2.0)
            + ((nonfolded_AR**0.5) * 2 * ((nonfolded_taper_ratio + 1.0)))
            * dNFTR_dTaperRatio
        )

        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Fuel.WING_FUEL_FRACTION] = (
            0.888889
            * tc_ratio_mean_folded
            * (nonfolded_wing_area**1.5)
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) / ((nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0))
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.THICKNESS_TO_CHORD_ROOT] = (
            fuel_vol_frac
            * 0.888889
            * dTCRMeanFolded_dTCR
            * (nonfolded_wing_area**1.5)
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) / ((nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0))
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.THICKNESS_TO_CHORD_TIP] = (
            fuel_vol_frac
            * 0.888889
            * dTCRMeanFolded_dTCT
            * (nonfolded_wing_area**1.5)
            * (2.0 * nonfolded_taper_ratio + 1.0)
        ) / ((nonfolded_AR**0.5) * ((nonfolded_taper_ratio + 1.0) ** 2.0))
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.SPAN] = (
            den * dNum_dWingspan - num * dDen_dWingspan
        ) / den**2
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.AREA] = (
            den * dNum_dWingArea - num * dDen_dWingArea
        ) / den**2
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, Aircraft.Wing.TAPER_RATIO] = (
            den * dNum_dTaperRatio - num * dDen_dTaperRatio
        ) / den**2
        J[Aircraft.Fuel.WING_VOLUME_GEOMETRIC_MAX, wrt] = (
            den * dNum_dy - num * dDen_dy) / den**2


class WingGroup(om.Group):
    def initialize(self):

        self.options.declare(
            'aviary_options', types=AviaryValues,
            desc='collection of Aircraft/Mission specific options'
        )

    def setup(self):

        aviary_options = self.options['aviary_options']

        size = self.add_subsystem(
            "size",
            WingSize(aviary_options=aviary_options,),
            promotes_inputs=["aircraft:*", "mission:*"],
            promotes_outputs=["aircraft:*"],
        )

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless') or self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            self.add_subsystem(
                "dimensionless_calcs",
                DimensionalNonDimensionalInterchange(aviary_options=aviary_options),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"]
            )

        parameters = self.add_subsystem(
            "parameters",
            WingParameters(aviary_options=aviary_options,),
            promotes_inputs=["aircraft:*"],
            promotes_outputs=["aircraft:*"],
        )

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_STRUT, units='unitless'):
            strut = self.add_subsystem(
                "strut",
                StrutGeom(
                    aviary_options=aviary_options,
                ),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"],
            )

        if self.options["aviary_options"].get_val(Aircraft.Wing.HAS_FOLD, units='unitless'):
            fold = self.add_subsystem(
                "fold",
                WingFold(
                    aviary_options=aviary_options,
                ),
                promotes_inputs=["aircraft:*"],
                promotes_outputs=["aircraft:*"],
            )

            if not self.options["aviary_options"].get_val(Aircraft.Wing.CHOOSE_FOLD_LOCATION, units='unitless'):
                check_fold_location_definition(None, aviary_options)
                self.promotes("strut", outputs=["strut_y"])
                self.promotes("fold", inputs=["strut_y"])

        self.set_input_defaults(Aircraft.Wing.ASPECT_RATIO, val=10.13, units="unitless")
        self.set_input_defaults(Aircraft.Wing.TAPER_RATIO, val=0.33, units="unitless")
        self.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_ROOT, val=0.11, units="unitless"
        )
        self.set_input_defaults(
            Aircraft.Wing.THICKNESS_TO_CHORD_TIP, val=0.1, units="unitless"
        )
